# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_context import context as oslo_context
from oslo_log import log as logging
from oslo_utils import timeutils
from oslo_utils import uuidutils

from senlin.common import consts
from senlin.common import context as senlin_context
from senlin.common import exception
from senlin.common import utils
from senlin.drivers import base as driver_base
from senlin.objects import credential as co
from senlin.objects import receiver as ro

LOG = logging.getLogger(__name__)


class Receiver(object):
    """Create a Receiver which is used to trigger a cluster action."""

    def __new__(cls, rtype, cluster_id=None, action=None, **kwargs):
        """Create a new receiver object.

        :param rtype: Type name of receiver.
        :param cluster_id: ID of the targeted cluster.
        :param action: Targeted action for execution.
        :param kwargs: A dict containing optional parameters.
        :returns: An instance of a specific sub-class of Receiver.
        """
        if rtype == consts.RECEIVER_WEBHOOK:
            from senlin.engine.receivers import webhook
            ReceiverClass = webhook.Webhook
        elif rtype == consts.RECEIVER_MESSAGE:
            from senlin.engine.receivers import message
            ReceiverClass = message.Message
        else:
            ReceiverClass = Receiver

        return super(Receiver, cls).__new__(ReceiverClass)

    def __init__(self, rtype, cluster_id=None, action=None, **kwargs):

        self.id = kwargs.get('id', None)
        self.name = kwargs.get('name', None)
        self.type = rtype
        self.user = kwargs.get('user', '')
        self.project = kwargs.get('project', '')
        self.domain = kwargs.get('domain', '')

        self.created_at = kwargs.get('created_at', None)
        self.updated_at = kwargs.get('updated_at', None)

        self.cluster_id = cluster_id
        self.action = action
        self.actor = kwargs.get('actor', {})
        self.params = kwargs.get('params', {})
        self.channel = kwargs.get('channel', {})

    def store(self, context, update=False):
        """Store the receiver in database and return its ID.

        :param context: Context for DB operations.
        """
        timestamp = timeutils.utcnow(True)
        self.created_at = timeutils.utcnow(True)
        values = {
            'id': self.id,
            'name': self.name,
            'type': self.type,
            'user': self.user,
            'project': self.project,
            'domain': self.domain,
            'created_at': self.created_at,
            'updated_at': self.updated_at,
            'cluster_id': self.cluster_id,
            'actor': self.actor,
            'action': self.action,
            'params': self.params,
            'channel': self.channel,
        }

        if update:
            self.updated_at = timestamp
            values['updated_at'] = timestamp
            ro.Receiver.update(context, self.id, values)
        else:
            self.created_at = timestamp
            values['created_at'] = timestamp
            receiver = ro.Receiver.create(context, values)
            self.id = receiver.id

        return self.id

    @classmethod
    def create(cls, context, rtype, cluster, action, **kwargs):
        cdata = dict()
        if rtype == consts.RECEIVER_WEBHOOK and context.is_admin:
            # use object owner if request is from admin
            cred = co.Credential.get(context, cluster.user, cluster.project)
            trust_id = cred['cred']['openstack']['trust']
            cdata['trust_id'] = trust_id
        else:
            # otherwise, use context user
            cdata['trust_id'] = context.trusts

        kwargs['actor'] = cdata
        kwargs['user'] = context.user_id
        kwargs['project'] = context.project_id
        kwargs['domain'] = context.domain_id
        kwargs['id'] = uuidutils.generate_uuid()
        cluster_id = cluster.id if cluster else None
        obj = cls(rtype, cluster_id, action, **kwargs)
        obj.initialize_channel(context)
        obj.store(context)

        return obj

    @classmethod
    def _from_object(cls, receiver):
        """Construct a receiver from receiver object.

        @param cls: The target class.
        @param receiver: a receiver object that contains all fields.
        """
        kwargs = {
            'id': receiver.id,
            'name': receiver.name,
            'user': receiver.user,
            'project': receiver.project,
            'domain': receiver.domain,
            'created_at': receiver.created_at,
            'updated_at': receiver.updated_at,
            'actor': receiver.actor,
            'params': receiver.params,
            'channel': receiver.channel,
        }

        return cls(receiver.type, receiver.cluster_id, receiver.action,
                   **kwargs)

    @classmethod
    def load(cls, context, receiver_id=None, receiver_obj=None,
             project_safe=True):
        """Retrieve a receiver from database.

        @param context: the context for db operations.
        @param receiver_id: the unique ID of the receiver to retrieve.
        @param receiver_obj: the DB object of a receiver to retrieve.
        @param project_safe: Optional parameter specifying whether only
                             receiver belong to the context.project will be
                             loaded.
        """
        if receiver_obj is None:
            receiver_obj = ro.Receiver.get(context, receiver_id,
                                           project_safe=project_safe)
            if receiver_obj is None:
                raise exception.ResourceNotFound(type='receiver',
                                                 id=receiver_id)

        return cls._from_object(receiver_obj)

    def to_dict(self):
        info = {
            'id': self.id,
            'name': self.name,
            'type': self.type,
            'user': self.user,
            'project': self.project,
            'domain': self.domain,
            'created_at': utils.isotime(self.created_at),
            'updated_at': utils.isotime(self.updated_at),
            'cluster_id': self.cluster_id,
            'actor': self.actor,
            'action': self.action,
            'params': self.params,
            'channel': self.channel,
        }
        return info

    def initialize_channel(self, context):
        return {}

    def release_channel(self, context):
        return

    def notify(self, context, params=None):
        return

    @classmethod
    def delete(cls, context, receiver_id):
        """Delete a receiver.

        @param context: the context for db operations.
        @param receiver_id: the unique ID of the receiver to delete.
        """
        receiver_obj = cls.load(context, receiver_id=receiver_id)
        receiver_obj.release_channel(context)
        ro.Receiver.delete(context, receiver_obj.id)

        return

    def _get_base_url(self):
        base = None
        service_cred = senlin_context.get_service_credentials()
        kc = driver_base.SenlinDriver().identity(service_cred)
        try:
            base = kc.get_senlin_endpoint()
        except exception.InternalError as ex:
            LOG.warning('Senlin endpoint can not be found: %s.', ex)

        return base

    def _build_conn_params(self, user, project):
        """Build connection params for specific user and project.

        :param user: The ID of the user for which a trust will be used.
        :param project: The ID of the project for which a trust will be used.
        :returns: A dict containing the required parameters for connection
                  creation.
        """
        service_creds = senlin_context.get_service_credentials()
        params = {
            'username': service_creds.get('username'),
            'password': service_creds.get('password'),
            'auth_url': service_creds.get('auth_url'),
            'user_domain_name': service_creds.get('user_domain_name')
        }

        cred = co.Credential.get(oslo_context.get_current(), user, project)
        if cred is None:
            raise exception.TrustNotFound(trustor=user)
        params['trust_id'] = cred.cred['openstack']['trust']

        return params



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import os
import six

from senlin.engine import parser
from senlin.tests.unit.common import base


class ParserTest(base.SenlinTestCase):

    json_template = """
    {
      "type": "os.heat.stack",
      "version": 1.0,
      "properties": {
        "name": "random_string_stack",
        "template": "random_string_stack.yaml"
      }
    }
    """

    yaml_template = """
    type: os.heat.stack
    version: 1.0
    properties:
      name: random_string_stack
      template: random_string_stack.yaml
    """

    expect_result = {
        "type": "os.heat.stack",
        "version": 1,
        "properties": {
            "name": "random_string_stack",
            "template": "random_string_stack.yaml"
        }
    }

    def test_parse_json_success(self):
        result = parser.simple_parse(self.json_template)
        self.assertEqual(self.expect_result, result)

    def test_parse_yaml_success(self):
        result = parser.simple_parse(self.yaml_template)
        self.assertEqual(self.expect_result, result)

    def test_parse_string(self):
        tmpl_str = 'json string'
        ex = self.assertRaises(ValueError,
                               parser.simple_parse,
                               tmpl_str)
        self.assertEqual('The input is not a JSON object or YAML mapping.',
                         six.text_type(ex))

    def test_parse_list(self):
        tmpl_str = '["foo" , "bar"]'
        ex = self.assertRaises(ValueError,
                               parser.simple_parse,
                               tmpl_str)
        self.assertEqual('The input is not a JSON object or YAML mapping.',
                         six.text_type(ex))

    def test_parse_invalid_yaml_and_json_template(self):
        tmpl_str = '{test'
        ex = self.assertRaises(ValueError,
                               parser.simple_parse,
                               tmpl_str)
        self.assertIn('Error parsing input:',
                      six.text_type(ex))


class ParseTemplateIncludeFiles(base.SenlinTestCase):
    scenarios = [
        ('include_from_file_without_path', dict(
            tmpl_str='foo: !include a.file',
            url_path='file:///tmp/a.file',
        )),
        ('include_from_file_with_path', dict(
            tmpl_str='foo: !include file:///tmp/a.file',
            url_path='file:///tmp/a.file',
        )),
        ('include_from_http', dict(
            tmpl_str='foo: !include http://tmp/a.file',
            url_path='http://tmp/a.file',
        )),
        ('include_from_https', dict(
            tmpl_str='foo: !include https://tmp/a.file',
            url_path='https://tmp/a.file',
        ))

    ]

    @mock.patch.object(six.moves.urllib.request, 'urlopen')
    @mock.patch.object(os.path, 'abspath')
    def test_parse_template(self, mock_abspath, mock_urlopen):
        fetch_data = 'bar'
        expect_result = {
            'foo': 'bar'
        }

        mock_abspath.return_value = '/tmp/a.file'
        mock_urlopen.side_effect = [
            six.moves.urllib.error.URLError('oops'),
            six.moves.cStringIO(fetch_data)
        ]

        ex = self.assertRaises(IOError,
                               parser.simple_parse,
                               self.tmpl_str)
        self.assertIn('Failed retrieving file %s:' % self.url_path,
                      six.text_type(ex))
        result = parser.simple_parse(self.tmpl_str)
        self.assertEqual(expect_result,
                         result)
        mock_urlopen.assert_has_calls([
            mock.call(self.url_path),
            mock.call(self.url_path)
        ])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from senlin.common import constraints
from senlin.common import consts
from senlin.common import exception as exc
from senlin.common.i18n import _
from senlin.common import scaleutils
from senlin.common import schema
from senlin.engine import health_manager
from senlin.policies import base


class HealthPolicy(base.Policy):
    """Policy for health management of a cluster."""

    VERSION = '1.0'
    VERSIONS = {
        '1.0': [
            {'status': consts.EXPERIMENTAL, 'since': '2017.02'},
            {'status': consts.SUPPORTED, 'since': '2018.06'},
        ]
    }
    PRIORITY = 600

    TARGET = [
        ('BEFORE', consts.CLUSTER_RECOVER),
        ('BEFORE', consts.CLUSTER_DEL_NODES),
        ('BEFORE', consts.CLUSTER_SCALE_IN),
        ('BEFORE', consts.CLUSTER_RESIZE),
        ('BEFORE', consts.NODE_DELETE),
        ('AFTER', consts.CLUSTER_DEL_NODES),
        ('AFTER', consts.CLUSTER_SCALE_IN),
        ('AFTER', consts.CLUSTER_RESIZE),
        ('AFTER', consts.NODE_DELETE),
    ]

    # Should be ANY if profile provides health check support?
    PROFILE_TYPE = [
        'os.nova.server',
        'os.heat.stack',
    ]

    KEYS = (DETECTION, RECOVERY) = ('detection', 'recovery')

    _DETECTION_KEYS = (
        DETECTION_TYPE, DETECTION_OPTIONS,
    ) = (
        'type', 'options'
    )

    _DETECTION_OPTIONS = (
        DETECTION_INTERVAL, POLL_URL, POLL_URL_SSL_VERIFY,
        POLL_URL_HEALTHY_RESPONSE, POLL_URL_RETRY_LIMIT,
        POLL_URL_RETRY_INTERVAL, NODE_UPDATE_TIMEOUT,
    ) = (
        'interval', 'poll_url', 'poll_url_ssl_verify',
        'poll_url_healthy_response', 'poll_url_retry_limit',
        'poll_url_retry_interval', 'node_update_timeout',
    )

    _RECOVERY_KEYS = (
        RECOVERY_ACTIONS, RECOVERY_FENCING, RECOVERY_DELETE_TIMEOUT,
        RECOVERY_FORCE_RECREATE,
    ) = (
        'actions', 'fencing', 'node_delete_timeout', 'node_force_recreate',
    )

    FENCING_OPTION_VALUES = (
        COMPUTE,
        # STORAGE, NETWORK,
    ) = (
        'COMPUTE',
        # 'STORAGE', 'NETWORK'
    )

    ACTION_KEYS = (
        ACTION_NAME, ACTION_PARAMS,
    ) = (
        'name', 'params',
    )

    properties_schema = {
        DETECTION: schema.Map(
            _('Policy aspect for node failure detection.'),
            schema={
                DETECTION_TYPE: schema.String(
                    _('Type of node failure detection.'),
                    constraints=[
                        constraints.AllowedValues(consts.DETECTION_TYPES),
                    ],
                    required=True,
                ),
                DETECTION_OPTIONS: schema.Map(
                    schema={
                        DETECTION_INTERVAL: schema.Integer(
                            _("Number of seconds between pollings. Only "
                              "required when type is 'NODE_STATUS_POLLING' or "
                              "'NODE_STATUS_POLL_URL'."),
                            default=60,
                        ),
                        POLL_URL: schema.String(
                            _("URL to poll for node status. See documentation "
                              "for valid expansion parameters. Only required "
                              "when type is 'NODE_STATUS_POLL_URL'."),
                            default='',
                        ),
                        POLL_URL_SSL_VERIFY: schema.Boolean(
                            _("Whether to verify SSL when calling URL to poll "
                              "for node status. Only required when type is "
                              "'NODE_STATUS_POLL_URL'."),
                            default=True,
                        ),
                        POLL_URL_HEALTHY_RESPONSE: schema.String(
                            _("String pattern in the poll URL response body "
                              "that indicates a healthy node. "
                              "Required when type is 'NODE_STATUS_POLL_URL'."),
                            default='',
                        ),
                        POLL_URL_RETRY_LIMIT: schema.Integer(
                            _("Number of times to retry URL polling when its "
                              "return body is missing "
                              "POLL_URL_HEALTHY_RESPONSE string before a node "
                              "is considered down. Required when type is "
                              "'NODE_STATUS_POLL_URL'."),
                            default=3,
                        ),
                        POLL_URL_RETRY_INTERVAL: schema.Integer(
                            _("Number of seconds between URL polling retries "
                              "before a node is considered down. "
                              "Required when type is 'NODE_STATUS_POLL_URL'."),
                            default=3,
                        ),
                        NODE_UPDATE_TIMEOUT: schema.Integer(
                            _("Number of seconds since last node update to "
                              "wait before checking node health. "
                              "Required when type is 'NODE_STATUS_POLL_URL'."),
                            default=300,
                        ),
                    },
                    default={}
                ),
            },
            required=True,
        ),

        RECOVERY: schema.Map(
            _('Policy aspect for node failure recovery.'),
            schema={
                RECOVERY_ACTIONS: schema.List(
                    _('List of actions to try for node recovery.'),
                    schema=schema.Map(
                        _('Action to try for node recovery.'),
                        schema={
                            ACTION_NAME: schema.String(
                                _("Name of action to execute."),
                                constraints=[
                                    constraints.AllowedValues(
                                        consts.RECOVERY_ACTIONS),
                                ],
                                required=True
                            ),
                            ACTION_PARAMS: schema.Map(
                                _("Parameters for the action")
                            ),
                        }
                    )
                ),
                RECOVERY_FENCING: schema.List(
                    _('List of services to be fenced.'),
                    schema=schema.String(
                        _('Service to be fenced.'),
                        constraints=[
                            constraints.AllowedValues(FENCING_OPTION_VALUES),
                        ],
                        required=True,
                    ),
                ),
                RECOVERY_DELETE_TIMEOUT: schema.Integer(
                    _("Number of seconds to wait for node deletion to "
                      "finish and start node creation for recreate "
                      "recovery option. Required when type is "
                      "'NODE_STATUS_POLL_URL and recovery action "
                      "is RECREATE'."),
                    default=20,
                ),
                RECOVERY_FORCE_RECREATE: schema.Boolean(
                    _("Whether to create node even if node deletion "
                      "failed. Required when type is "
                      "'NODE_STATUS_POLL_URL' and action recovery "
                      "action is RECREATE."),
                    default=False,
                ),
            }
        ),
    }

    def __init__(self, name, spec, **kwargs):
        super(HealthPolicy, self).__init__(name, spec, **kwargs)

        self.check_type = self.properties[self.DETECTION][self.DETECTION_TYPE]

        options = self.properties[self.DETECTION][self.DETECTION_OPTIONS]
        self.interval = options.get(self.DETECTION_INTERVAL, 60)
        self.poll_url = options.get(self.POLL_URL, '')
        self.poll_url_ssl_verify = options.get(self.POLL_URL_SSL_VERIFY, True)
        self.poll_url_healthy_response = options.get(
            self.POLL_URL_HEALTHY_RESPONSE, '')
        self.poll_url_retry_limit = options.get(self.POLL_URL_RETRY_LIMIT, '')
        self.poll_url_retry_interval = options.get(
            self.POLL_URL_RETRY_INTERVAL, '')
        self.node_update_timeout = options.get(self.NODE_UPDATE_TIMEOUT, 300)

        recover_settings = self.properties[self.RECOVERY]
        self.recover_actions = recover_settings[self.RECOVERY_ACTIONS]
        self.fencing_types = recover_settings[self.RECOVERY_FENCING]
        self.node_delete_timeout = recover_settings.get(
            self.RECOVERY_DELETE_TIMEOUT, None)
        self.node_force_recreate = recover_settings.get(
            self.RECOVERY_FORCE_RECREATE, False)

    def validate(self, context, validate_props=False):
        super(HealthPolicy, self).validate(context,
                                           validate_props=validate_props)

        if len(self.recover_actions) > 1:
            message = _("Only one '%s' is supported for now."
                        ) % self.RECOVERY_ACTIONS
            raise exc.ESchema(message=message)

        if self.interval < cfg.CONF.health_check_interval_min:
            message = _("Specified interval of %(interval)d seconds has to be "
                        "larger than health_check_interval_min of "
                        "%(min_interval)d seconds set in configuration."
                        ) % {"interval": self.interval,
                             "min_interval":
                                 cfg.CONF.health_check_interval_min}
            raise exc.InvalidSpec(message=message)

        # TODO(Qiming): Add detection of duplicated action names when
        # support to list of actions is implemented.

    def attach(self, cluster, enabled=True):
        """"Hook for policy attach.

        Register the cluster for health management.

        :param cluster: The cluster to which the policy is being attached to.
        :param enabled: The attached cluster policy is enabled or disabled.
        :return: A tuple comprising execution result and policy data.
        """
        p_type = cluster.rt['profile'].type_name
        action_names = [a['name'] for a in self.recover_actions]
        if p_type != 'os.nova.server':
            if consts.RECOVER_REBUILD in action_names:
                err_msg = _("Recovery action REBUILD is only applicable to "
                            "os.nova.server clusters.")
                return False, err_msg

            if consts.RECOVER_REBOOT in action_names:
                err_msg = _("Recovery action REBOOT is only applicable to "
                            "os.nova.server clusters.")
                return False, err_msg

        kwargs = {
            'check_type': self.check_type,
            'interval': self.interval,
            'params': {
                'recover_action': self.recover_actions,
                'poll_url': self.poll_url,
                'poll_url_ssl_verify': self.poll_url_ssl_verify,
                'poll_url_healthy_response': self.poll_url_healthy_response,
                'poll_url_retry_limit': self.poll_url_retry_limit,
                'poll_url_retry_interval': self.poll_url_retry_interval,
                'node_update_timeout': self.node_update_timeout,
                'node_delete_timeout': self.node_delete_timeout,
                'node_force_recreate': self.node_force_recreate,
            },
            'enabled': enabled
        }

        health_manager.register(cluster.id, engine_id=None, **kwargs)

        data = {
            'check_type': self.check_type,
            'interval': self.interval,
            'poll_url': self.poll_url,
            'poll_url_ssl_verify': self.poll_url_ssl_verify,
            'poll_url_healthy_response': self.poll_url_healthy_response,
            'poll_url_retry_limit': self.poll_url_retry_limit,
            'poll_url_retry_interval': self.poll_url_retry_interval,
            'node_update_timeout': self.node_update_timeout,
            'node_delete_timeout': self.node_delete_timeout,
            'node_force_recreate': self.node_force_recreate,
        }

        return True, self._build_policy_data(data)

    def detach(self, cluster):
        """Hook for policy detach.

        Unregister the cluster for health management.
        :param cluster: The target cluster.
        :returns: A tuple comprising the execution result and reason.
        """
        health_manager.unregister(cluster.id)
        return True, ''

    def pre_op(self, cluster_id, action, **args):
        """Hook before action execution.

        One of the task for this routine is to disable health policy if the
        action is a request that will shrink the cluster. The reason is that
        the policy may attempt to recover nodes that are to be deleted.

        :param cluster_id: The ID of the target cluster.
        :param action: The action to be examined.
        :param kwargs args: Other keyword arguments to be checked.
        :returns: Boolean indicating whether the checking passed.
        """
        if action.action in (consts.CLUSTER_SCALE_IN,
                             consts.CLUSTER_DEL_NODES,
                             consts.NODE_DELETE):
            health_manager.disable(cluster_id)
            return True

        if action.action == consts.CLUSTER_RESIZE:
            deletion = action.data.get('deletion', None)
            if deletion:
                health_manager.disable(cluster_id)
                return True

            cluster = action.entity
            current = len(cluster.nodes)
            res, reason = scaleutils.parse_resize_params(action, cluster,
                                                         current)
            if res == base.CHECK_ERROR:
                action.data['status'] = base.CHECK_ERROR
                action.data['reason'] = reason
                return False

            if action.data.get('deletion', None):
                health_manager.disable(cluster_id)
                return True

        pd = {
            'recover_action': self.recover_actions,
            'fencing': self.fencing_types,
        }
        action.data.update({'health': pd})
        action.store(action.context)

        return True

    def post_op(self, cluster_id, action, **args):
        """Hook before action execution.

        One of the task for this routine is to re-enable health policy if the
        action is a request that will shrink the cluster thus the policy has
        been temporarily disabled.

        :param cluster_id: The ID of the target cluster.
        :param action: The action to be examined.
        :param kwargs args: Other keyword arguments to be checked.
        :returns: Boolean indicating whether the checking passed.
        """
        if action.action in (consts.CLUSTER_SCALE_IN,
                             consts.CLUSTER_DEL_NODES,
                             consts.NODE_DELETE):
            health_manager.enable(cluster_id)
            return True

        if action.action == consts.CLUSTER_RESIZE:
            deletion = action.data.get('deletion', None)
            if deletion:
                health_manager.enable(cluster_id)
                return True

            cluster = action.entity
            current = len(cluster.nodes)
            res, reason = scaleutils.parse_resize_params(action, cluster,
                                                         current)
            if res == base.CHECK_ERROR:
                action.data['status'] = base.CHECK_ERROR
                action.data['reason'] = reason
                return False

            if action.data.get('deletion', None):
                health_manager.enable(cluster_id)
                return True

        return True



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
from oslo_utils import uuidutils

from senlin.drivers.os import neutron_v2
from senlin.drivers import sdk
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils


class TestNeutronV2Driver(base.SenlinTestCase):

    def setUp(self):
        super(TestNeutronV2Driver, self).setUp()
        self.context = utils.dummy_context()
        self.conn_params = self.context.to_dict()
        self.conn = mock.Mock()
        with mock.patch.object(sdk, 'create_connection') as mock_creare_conn:
            mock_creare_conn.return_value = self.conn
            self.nc = neutron_v2.NeutronClient(self.context)

    @mock.patch.object(sdk, 'create_connection')
    def test_init(self, mock_create_connection):
        params = self.conn_params
        neutron_v2.NeutronClient(params)
        mock_create_connection.assert_called_once_with(params)

    def test_network_get_with_uuid(self):
        net_id = uuidutils.generate_uuid()
        network_obj = mock.Mock()

        self.conn.network.find_network.return_value = network_obj
        res = self.nc.network_get(net_id)
        self.conn.network.find_network.assert_called_once_with(net_id, False)
        self.assertEqual(network_obj, res)

    def test_network_get_with_name(self):
        net_id = 'network_identifier'
        net1 = mock.Mock()
        net2 = mock.Mock()
        self.conn.network.networks.return_value = [net1, net2]

        res = self.nc.network_get(net_id)
        self.assertEqual(0, self.conn.network.find_network.call_count)
        self.conn.network.networks.assert_called_once_with(name=net_id)
        self.assertEqual(net1, res)

    def test_port_find(self):
        port_id = 'port_identifier'
        port_obj = mock.Mock()

        self.conn.network.find_port.return_value = port_obj
        res = self.nc.port_find(port_id)
        self.conn.network.find_port.assert_called_once_with(port_id, False)
        self.assertEqual(port_obj, res)

    def test_security_group_find(self):
        sg_id = 'sg_identifier'
        sg_obj = mock.Mock()

        self.conn.network.find_security_group.return_value = sg_obj
        res = self.nc.security_group_find(sg_id)
        self.conn.network.find_security_group.assert_called_once_with(
            sg_id, False)
        self.assertEqual(sg_obj, res)

    def test_subnet_get(self):
        subnet_id = 'subnet_identifier'
        subnet_obj = mock.Mock()

        self.conn.network.find_subnet.return_value = subnet_obj
        res = self.nc.subnet_get(subnet_id)
        self.conn.network.find_subnet.assert_called_once_with(subnet_id, False)
        self.assertEqual(subnet_obj, res)

    def test_port_create(self):
        port_attr = {
            'network_id': 'foo'
        }
        self.nc.port_create(**port_attr)
        self.conn.network.create_port.assert_called_once_with(
            network_id='foo')

    def test_port_delete(self):
        self.nc.port_delete(port='foo')
        self.conn.network.delete_port.assert_called_once_with(
            port='foo', ignore_missing=True)

    def test_port_update(self):
        attr = {
            'name': 'new_name'
        }
        self.nc.port_update('fake_port', **attr)
        self.conn.network.update_port.assert_called_once_with(
            'fake_port', **attr)

    def test_floatingip_find(self):
        floatingip_id = 'fake_id'
        fip_obj = mock.Mock()

        self.conn.network.find_ip.return_value = fip_obj
        res = self.nc.floatingip_find(floatingip_id)
        self.conn.network.find_ip.assert_called_once_with(
            floatingip_id, ignore_missing=False)
        self.assertEqual(fip_obj, res)

    def test_floatingip_list_by_port_id(self):
        port_id = 'port_id'
        fip_obj_iter = iter([mock.Mock()])

        self.conn.network.ips.return_value = fip_obj_iter
        res = self.nc.floatingip_list(port=port_id)
        self.conn.network.ips.assert_called_once_with(port_id=port_id)
        self.assertEqual(1, len(res))

    def test_floatingip_create(self):
        attr = {
            'network_id': 'foo'
        }
        self.nc.floatingip_create(**attr)
        self.conn.network.create_ip.assert_called_once_with(
            network_id='foo')

    def test_floatingip_delete(self):
        self.nc.floatingip_delete(floating_ip='foo')
        self.conn.network.delete_ip.assert_called_once_with(
            'foo', ignore_missing=True)

    def test_floatingip_update(self):
        attr = {
            'port_id': 'fake_port'
        }
        self.nc.floatingip_update('fake_floatingip', **attr)
        self.conn.network.update_ip.assert_called_once_with(
            'fake_floatingip', **attr)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
from oslo_context import context as oslo_context
import six

from senlin.common import consts
from senlin.common import exception as exc
from senlin.common import scaleutils
from senlin.drivers import base as driver_base
from senlin.engine import cluster_policy
from senlin.objects import node as no
from senlin.policies import base as policy_base
from senlin.policies import lb_policy
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils


class TestLoadBalancingPolicy(base.SenlinTestCase):

    def setUp(self):
        super(TestLoadBalancingPolicy, self).setUp()
        self.context = utils.dummy_context()
        self.spec = {
            'type': 'senlin.policy.loadbalance',
            'version': '1.0',
            'properties': {
                'pool': {
                    'id': '',
                    'protocol': 'HTTP',
                    'protocol_port': 80,
                    'subnet': 'internal-subnet',
                    'lb_method': 'ROUND_ROBIN',
                    'admin_state_up': True,
                    'session_persistence': {
                        'type': 'SOURCE_IP',
                        'cookie_name': 'whatever'
                    }
                },
                'vip': {
                    'address': '192.168.1.100',
                    'subnet': 'external-subnet',
                    'connection_limit': 500,
                    'protocol': 'HTTP',
                    'protocol_port': 80,
                    'admin_state_up': True,
                },
                'health_monitor': {
                    'type': 'HTTP',
                    'delay': 10,
                    'timeout': 5,
                    'max_retries': 3,
                    'admin_state_up': True,
                    'http_method': 'GET',
                    'url_path': '/index.html',
                    'expected_codes': '200,201,202'
                },
                'lb_status_timeout': 300
            }
        }
        self.sd = mock.Mock()
        self.patchobject(driver_base, 'SenlinDriver', return_value=self.sd)
        self.lb_driver = mock.Mock()
        self.net_driver = mock.Mock()
        self.octavia_driver = mock.Mock()

    @mock.patch.object(lb_policy.LoadBalancingPolicy, 'validate')
    def test_init(self, mock_validate):
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)

        self.assertIsNone(policy.id)
        self.assertEqual('test-policy', policy.name)
        self.assertEqual('senlin.policy.loadbalance-1.0', policy.type)
        self.assertEqual(self.spec['properties']['pool'], policy.pool_spec)
        self.assertEqual(self.spec['properties']['vip'], policy.vip_spec)
        self.assertIsNone(policy.lb)

    def test_init_with_default_value(self):
        spec = {
            'type': 'senlin.policy.loadbalance',
            'version': '1.0',
            'properties': {
                'pool': {'subnet': 'internal-subnet'},
                'vip': {'subnet': 'external-subnet'}
            }
        }
        default_spec = {
            'type': 'senlin.policy.loadbalance',
            'version': '1.0',
            'properties': {
                'pool': {
                    'id': None,
                    'protocol': 'HTTP',
                    'protocol_port': 80,
                    'subnet': 'internal-subnet',
                    'lb_method': 'ROUND_ROBIN',
                    'admin_state_up': True,
                    'session_persistence': {},
                },
                'vip': {
                    'address': None,
                    'subnet': 'external-subnet',
                    'connection_limit': -1,
                    'protocol': 'HTTP',
                    'protocol_port': 80,
                    'admin_state_up': True,
                },
                'lb_status_timeout': 300
            }
        }

        policy = lb_policy.LoadBalancingPolicy('test-policy', spec)

        self.assertIsNone(policy.id)
        self.assertEqual('test-policy', policy.name)
        self.assertEqual('senlin.policy.loadbalance-1.0', policy.type)
        self.assertEqual(default_spec['properties']['pool'], policy.pool_spec)
        self.assertEqual(default_spec['properties']['vip'], policy.vip_spec)
        self.assertEqual(default_spec['properties']['lb_status_timeout'],
                         policy.lb_status_timeout)
        self.assertIsNone(policy.lb)

    def test_loadbalancer_value(self):
        spec = {
            'type': 'senlin.policy.loadbalance',
            'version': '1.0',
            'properties': {
                'loadbalancer': 'LB_ID',
                'pool': {
                    'id': 'POOL_ID',
                    'subnet': 'internal-subnet'
                },
                'vip': {
                    'address': '192.168.1.100',
                    'subnet': 'external-subnet'
                },
                'health_monitor': {
                    'id': 'HM_ID'
                }
            }
        }
        self.spec['properties']['pool']['id'] = 'POOL_ID'
        self.spec['properties']['health_monitor']['id'] = 'HM_ID'
        self.spec['properties']['loadbalancer'] = 'LB_ID'
        self.spec['properties']['pool']['session_persistence'] = {}
        self.spec['properties']['vip']['connection_limit'] = -1
        policy = lb_policy.LoadBalancingPolicy('test-policy', spec)
        self.assertIsNone(policy.id)
        self.assertEqual('test-policy', policy.name)
        self.assertEqual('senlin.policy.loadbalance-1.0', policy.type)
        self.assertEqual(self.spec['properties']['pool'], policy.pool_spec)
        self.assertEqual(self.spec['properties']['vip'], policy.vip_spec)
        self.assertEqual(self.spec['properties']['loadbalancer'], policy.lb)

    @mock.patch.object(policy_base.Policy, 'validate')
    def test_validate_shallow(self, mock_validate):
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        ctx = mock.Mock()

        res = policy.validate(ctx, False)

        self.assertTrue(res)
        mock_validate.assert_called_with(ctx, False)

    @mock.patch.object(policy_base.Policy, 'validate')
    def test_validate_pool_subnet_notfound(self, mock_validate):
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._networkclient = self.net_driver
        policy._octaviaclient = self.octavia_driver
        ctx = mock.Mock(user='user1', project='project1')
        self.net_driver.subnet_get = mock.Mock(
            side_effect=exc.InternalError(code='404', message='not found'))

        ex = self.assertRaises(exc.InvalidSpec, policy.validate, ctx, True)

        mock_validate.assert_called_with(ctx, True)
        self.net_driver.subnet_get.assert_called_once_with('internal-subnet')
        self.assertEqual("The specified subnet 'internal-subnet' could not "
                         "be found.", six.text_type(ex))

    @mock.patch.object(policy_base.Policy, 'validate')
    def test_validate_vip_subnet_notfound(self, mock_validate):
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._networkclient = self.net_driver
        policy._octaviaclient = self.octavia_driver
        ctx = mock.Mock(user='user1', project='project1')
        self.net_driver.subnet_get = mock.Mock(
            side_effect=[
                mock.Mock(),  # for the internal (pool) one
                exc.InternalError(code='404', message='not found')
            ]
        )

        ex = self.assertRaises(exc.InvalidSpec, policy.validate, ctx, True)

        mock_validate.assert_called_with(ctx, True)
        self.net_driver.subnet_get.assert_has_calls([
            mock.call('internal-subnet'), mock.call('external-subnet')
        ])
        self.assertEqual("The specified subnet 'external-subnet' could not "
                         "be found.", six.text_type(ex))

    @mock.patch.object(policy_base.Policy, 'validate')
    def test_validate_loadbalancer_notfund(self, mock_validate):
        self.spec['properties']['loadbalancer'] = "LB_ID"
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._networkclient = self.net_driver
        policy._octaviaclient = self.octavia_driver
        ctx = mock.Mock(user='user1', project='project1')
        self.octavia_driver.loadbalancer_get = mock.Mock(
            side_effect=exc.InternalError(code='404', message='not found'))

        ex = self.assertRaises(exc.InvalidSpec, policy.validate, ctx, True)

        mock_validate.assert_called_with(ctx, True)
        self.octavia_driver.loadbalancer_get.assert_called_once_with('LB_ID')
        self.assertEqual("The specified loadbalancer 'LB_ID' could not "
                         "be found.", six.text_type(ex))

    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_build_policy_data')
    @mock.patch.object(policy_base.Policy, 'attach')
    @mock.patch.object(no.Node, 'update')
    def test_attach_succeeded(self, m_update, m_attach, m_build):
        cluster = mock.Mock(id='CLUSTER_ID', data={})
        node1 = mock.Mock(id='fake1', data={})
        node2 = mock.Mock(id='fake2', data={})
        cluster.nodes = [node1, node2]
        m_attach.return_value = (True, None)
        m_build.return_value = 'policy_data'
        data = {
            'loadbalancer': 'LB_ID',
            'vip_address': '192.168.1.100',
            'pool': 'POOL_ID'
        }
        self.lb_driver.lb_create.return_value = (True, data)
        self.lb_driver.member_add.side_effect = ['MEMBER1_ID', 'MEMBER2_ID']

        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy.id = 'FAKE_ID'
        policy._lbaasclient = self.lb_driver

        res, data = policy.attach(cluster)

        self.assertTrue(res)
        self.assertEqual('policy_data', data)
        self.lb_driver.lb_create.assert_called_once_with(policy.vip_spec,
                                                         policy.pool_spec,
                                                         policy.hm_spec)
        member_add_calls = [
            mock.call(node1, 'LB_ID', 'POOL_ID', 80, 'internal-subnet'),
            mock.call(node2, 'LB_ID', 'POOL_ID', 80, 'internal-subnet')
        ]
        self.lb_driver.member_add.assert_has_calls(member_add_calls)
        node_update_calls = [
            mock.call(mock.ANY, node1.id,
                      {'data': {'lb_member': 'MEMBER1_ID'}}),
            mock.call(mock.ANY, node2.id,
                      {'data': {'lb_member': 'MEMBER2_ID'}})
        ]
        m_update.assert_has_calls(node_update_calls)
        expected = {
            policy.id: {'vip_address': '192.168.1.100'}
        }
        self.assertEqual(expected, cluster.data['loadbalancers'])

    @mock.patch.object(policy_base.Policy, 'attach')
    def test_attach_failed_base_return_false(self, mock_attach):
        cluster = mock.Mock()
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        mock_attach.return_value = (False, 'data')

        res, data = policy.attach(cluster)

        self.assertFalse(res)
        self.assertEqual('data', data)

    @mock.patch.object(policy_base.Policy, 'attach')
    def test_attach_failed_lb_creation_error(self, m_attach):
        cluster = mock.Mock()
        m_attach.return_value = (True, None)

        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver

        # lb_driver.lb_create return False
        self.lb_driver.lb_create.return_value = (False, 'error')
        res = policy.attach(cluster)
        self.assertEqual((False, 'error'), res)

    @mock.patch.object(policy_base.Policy, 'attach')
    def test_attach_failed_member_add(self, mock_attach):
        cluster = mock.Mock()
        cluster.nodes = [mock.Mock(id='fake1'), mock.Mock(id='fake2')]
        mock_attach.return_value = (True, None)
        lb_data = {
            'loadbalancer': 'LB_ID',
            'vip_address': '192.168.1.100',
            'pool': 'POOL_ID'
        }
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver
        # lb_driver.member_add return None
        self.lb_driver.lb_create.return_value = (True, lb_data)
        self.lb_driver.member_add.return_value = None

        res = policy.attach(cluster)

        self.assertEqual((False, 'Failed in adding node into lb pool'), res)
        self.lb_driver.lb_delete.assert_called_once_with(**lb_data)

    def test_post_candidates_node_recover_reboot(self):
        node = mock.Mock(id='NODE1_ID')
        action = mock.Mock(action=consts.NODE_RECOVER)
        action.entity = node
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)

        candidates = policy._get_post_candidates(action)

        self.assertEqual(['NODE1_ID'], candidates)

    def test_post_candidates_node_recover_empty(self):
        node = mock.Mock(id='NODE1_ID')
        action = mock.Mock(action=consts.NODE_RECOVER,
                           outputs={})
        action.entity = node
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)

        candidates = policy._get_post_candidates(action)

        self.assertEqual(['NODE1_ID'], candidates)

    def test_post_candidates_cluster_resize(self):
        action = mock.Mock(action=consts.CLUSTER_RESIZE,
                           data={
                               'creation': {
                                   'nodes': ['NODE1_ID', 'NODE2_ID']
                               }
                           })
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)

        candidates = policy._get_post_candidates(action)

        self.assertEqual(['NODE1_ID', 'NODE2_ID'], candidates)

    def test_get_delete_candidates_for_node_delete(self):
        action = mock.Mock(action=consts.NODE_DELETE, inputs={}, data={},
                           entity=mock.Mock(id='NODE_ID'))
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)

        res = policy._get_delete_candidates('CLUSTERID', action)

        self.assertEqual(['NODE_ID'], res)

    def test_get_delete_candidates_no_deletion_data_del_nodes(self):
        action = mock.Mock(action=consts.CLUSTER_DEL_NODES, data={},
                           inputs={'candidates': ['node1', 'node2']})
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)

        res = policy._get_delete_candidates('CLUSTERID', action)

        self.assertEqual(['node1', 'node2'], res)

    @mock.patch.object(scaleutils, 'nodes_by_random')
    def test_get_delete_candidates_no_deletion_data_scale_in(self,
                                                             m_nodes_random):
        self.context = utils.dummy_context()
        node1 = mock.Mock(id='node1')
        node2 = mock.Mock(id='node2')
        node3 = mock.Mock(id='node3')
        cluster = mock.Mock()
        cluster.nodes = [node1, node2, node3]
        action = mock.Mock(action=consts.CLUSTER_SCALE_IN, data={})
        action.entity = cluster

        m_nodes_random.return_value = ['node1', 'node3']
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)

        res = policy._get_delete_candidates('CLUSTERID', action)

        m_nodes_random.assert_called_once_with([node1, node2, node3], 1)
        self.assertEqual(['node1', 'node3'], res)

    @mock.patch.object(scaleutils, 'parse_resize_params')
    @mock.patch.object(scaleutils, 'nodes_by_random')
    def test_get_delete_candidates_no_deletion_data_resize(
            self, m_nodes_random, m_parse_param):

        def _parse_param(action, cluster, current):
            action.data = {'deletion': {'count': 2}}

        self.context = utils.dummy_context()
        node1 = mock.Mock(id='node1')
        node2 = mock.Mock(id='node2')
        node3 = mock.Mock(id='node3')
        cluster = mock.Mock(id='cluster1')
        cluster.nodes = [node1, node2, node3]
        action = mock.Mock(action=consts.CLUSTER_RESIZE, data={})
        action.entity = cluster

        m_parse_param.side_effect = _parse_param
        m_nodes_random.return_value = ['node1', 'node3']

        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)

        res = policy._get_delete_candidates('CLUSTERID', action)

        m_parse_param.assert_called_once_with(action, cluster, 3)
        m_nodes_random.assert_called_once_with([node1, node2, node3], 2)
        self.assertEqual(['node1', 'node3'], res)

    @mock.patch.object(scaleutils, 'nodes_by_random')
    def test_get_delete_candidates_deletion_no_candidates(self,
                                                          m_nodes_random):
        self.context = utils.dummy_context()
        node1 = mock.Mock(id='node1')
        node2 = mock.Mock(id='node2')
        node3 = mock.Mock(id='node3')
        cluster = mock.Mock(id='cluster1')
        cluster.nodes = [node1, node2, node3]
        action = mock.Mock(action=consts.CLUSTER_RESIZE, data={})
        action.entity = cluster
        action.data = {'deletion': {'count': 1}}

        m_nodes_random.return_value = ['node2']

        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)

        res = policy._get_delete_candidates('CLUSTERID', action)

        m_nodes_random.assert_called_once_with([node1, node2, node3], 1)

        self.assertEqual(['node2'], res)
        self.assertEqual({'deletion': {'count': 1, 'candidates': ['node2']}},
                         action.data)

    def test_get_delete_candidates_deletion_count_is_zero(self):
        self.context = utils.dummy_context()
        action = mock.Mock(data={'deletion': {'number': 3}})
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)

        res = policy._get_delete_candidates('CLUSTERID', action)

        self.assertEqual([], res)

    @mock.patch.object(scaleutils, 'nodes_by_random')
    def test_get_delete_candidates_deletion_count_over_size(self,
                                                            m_nodes_random):
        node1 = mock.Mock(id='node1')
        node2 = mock.Mock(id='node2')
        node3 = mock.Mock(id='node3')
        cluster = mock.Mock(id='cluster1')
        cluster.nodes = [node1, node2, node3]
        action = mock.Mock(action=consts.CLUSTER_RESIZE, data={})
        action.entity = cluster
        action.data = {'deletion': {'count': 4}}

        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)

        policy._get_delete_candidates('CLUSTERID', action)

        m_nodes_random.assert_called_once_with([node1, node2, node3], 3)

    def test_get_delete_candidates_deletion_with_candidates(self):
        action = mock.Mock()
        action.data = {'deletion': {'count': 1, 'candidates': ['node3']}}

        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)

        res = policy._get_delete_candidates('CLUSTERID', action)
        self.assertEqual(['node3'], res)


@mock.patch.object(cluster_policy.ClusterPolicy, 'load')
@mock.patch.object(lb_policy.LoadBalancingPolicy, '_extract_policy_data')
class TestLoadBalancingPolicyOperations(base.SenlinTestCase):

    def setUp(self):
        super(TestLoadBalancingPolicyOperations, self).setUp()

        self.context = utils.dummy_context()
        self.spec = {
            'type': 'senlin.policy.loadbalance',
            'version': '1.0',
            'properties': {
                'pool': {
                    'protocol': 'HTTP',
                    'protocol_port': 80,
                    'subnet': 'test-subnet',
                    'lb_method': 'ROUND_ROBIN',
                    'admin_state_up': True,
                    'session_persistence': {
                        'type': 'SOURCE_IP',
                        'cookie_name': 'whatever'
                    }
                },
                'vip': {
                    'address': '192.168.1.100',
                    'subnet': 'test-subnet',
                    'connection_limit': 500,
                    'protocol': 'HTTP',
                    'protocol_port': 80,
                    'admin_state_up': True,
                },
                'health_monitor': {
                    'type': 'HTTP',
                    'delay': '1',
                    'timeout': 1,
                    'max_retries': 5,
                    'admin_state_up': True,
                    'http_method': 'GET',
                    'url_path': '/index.html',
                    'expected_codes': '200,201,202'
                }
            }
        }
        self.lb_driver = mock.Mock()
        self.patchobject(oslo_context, 'get_current')

    def test_detach_no_policy_data(self, m_extract, m_load):
        cluster = mock.Mock()
        m_extract.return_value = None
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver

        res, data = policy.detach(cluster)

        self.assertTrue(res)
        self.assertEqual('LB resources deletion succeeded.', data)

    def test_detach_succeeded(self, m_extract, m_load):
        cp = mock.Mock()
        policy_data = {
            'loadbalancer': 'LB_ID',
            'listener': 'LISTENER_ID',
            'pool': 'POOL_ID',
            'healthmonitor': 'HM_ID'
        }
        cp_data = {
            'LoadBalancingPolicy': {
                'version': '1.0',
                'data': policy_data
            }
        }
        cp.data = cp_data
        m_load.return_value = cp
        m_extract.return_value = policy_data
        self.lb_driver.lb_delete.return_value = (True, 'lb_delete succeeded.')
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver
        cluster = mock.Mock(
            id='CLUSTER_ID',
            data={
                'loadbalancers': {
                    policy.id: {'vip_address': '192.168.1.100'}

                }
            })
        node = mock.Mock(id='fake', data={})
        cluster.nodes = [node]

        res, data = policy.detach(cluster)

        self.assertTrue(res)
        self.assertEqual('lb_delete succeeded.', data)
        m_load.assert_called_once_with(mock.ANY, cluster.id, policy.id)
        m_extract.assert_called_once_with(cp_data)
        self.lb_driver.lb_delete.assert_called_once_with(**policy_data)
        self.assertEqual({}, cluster.data)

    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_remove_member')
    def test_detach_existed_lbass_succeeded(self, m_remove, m_extract,
                                            m_load):
        cp = mock.Mock()
        policy_data = {
            'loadbalancer': 'LB_ID',
            'listener': 'LISTENER_ID',
            'pool': 'POOL_ID',
            'healthmonitor': 'HM_ID',
            'preexisting': True,
        }
        cp_data = {
            'LoadBalancingPolicy': {
                'version': '1.0',
                'data': policy_data
            }
        }
        cp.data = cp_data
        m_load.return_value = cp
        m_extract.return_value = policy_data

        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver
        cluster = mock.Mock(
            id='CLUSTER_ID',
            data={
                'loadbalancers': {
                    policy.id: {'vip_address': '192.168.1.100'}

                }
            })
        node = mock.Mock(id='fake', data={})
        cluster.nodes = [node]
        m_remove.return_value = []

        res, data = policy.detach(cluster)

        self.assertTrue(res)
        self.assertEqual('LB resources deletion succeeded.', data)
        m_load.assert_called_once_with(mock.ANY, cluster.id, policy.id)
        m_extract.assert_called_once_with(cp_data)
        m_remove.assert_called_with(mock.ANY, ['fake'], cp, self.lb_driver)
        self.assertEqual({}, cluster.data)

    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_remove_member')
    def test_detach_existed_lbass_failed(self, m_remove, m_extract, m_load):
        cp = mock.Mock()
        policy_data = {
            'loadbalancer': 'LB_ID',
            'listener': 'LISTENER_ID',
            'pool': 'POOL_ID',
            'healthmonitor': 'HM_ID',
            'preexisting': True,
        }
        cp_data = {
            'LoadBalancingPolicy': {
                'version': '1.0',
                'data': policy_data
            }
        }
        cp.data = cp_data
        m_load.return_value = cp
        m_extract.return_value = policy_data

        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver
        cluster = mock.Mock(
            id='CLUSTER_ID',
            data={
                'loadbalancers': {
                    policy.id: {'vip_address': '192.168.1.100'}

                }
            })
        node1 = mock.Mock(id='node1', data={})
        node2 = mock.Mock(id='node2', data={})
        cluster.nodes = [node1, node2]
        m_remove.return_value = [node2.id]

        res, data = policy.detach(cluster)

        self.assertFalse(res)
        self.assertEqual('Failed to remove servers from existed LB.', data)
        m_load.assert_called_once_with(mock.ANY, cluster.id, policy.id)
        m_extract.assert_called_once_with(cp_data)
        m_remove.assert_called_with(mock.ANY, ['node1', 'node2'], cp,
                                    self.lb_driver)
        self.assertEqual({
            'loadbalancers': {
                None: {'vip_address': '192.168.1.100'}
            }},
            cluster.data)

    def test_detach_failed_lb_delete(self, m_extract, m_load):
        cluster = mock.Mock()
        policy_data = {
            'preexisting': False,
        }
        m_extract.return_value = policy_data
        self.lb_driver.lb_delete.return_value = (False, 'lb_delete failed.')

        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver

        res, data = policy.detach(cluster)

        self.assertFalse(res)
        self.assertEqual('lb_delete failed.', data)

    def test_post_op_no_nodes(self, m_extract, m_load):
        action = mock.Mock(data={})

        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)

        res = policy.post_op('FAKE_ID', action)

        self.assertIsNone(res)

    @mock.patch.object(no.Node, 'get')
    @mock.patch.object(no.Node, 'update')
    def test__add_member(self, m_node_update, m_node_get,
                         m_extract, m_load):
        node1 = mock.Mock(id='NODE1_ID', data={})
        node2 = mock.Mock(id='NODE2_ID', data={})
        action = mock.Mock(context='action_context',
                           action=consts.CLUSTER_RESIZE,
                           data={
                               'creation': {
                                   'nodes': ['NODE1_ID', 'NODE2_ID']
                               }
                           })
        cp = mock.Mock()
        policy_data = {
            'loadbalancer': 'LB_ID',
            'listener': 'LISTENER_ID',
            'pool': 'POOL_ID',
            'healthmonitor': 'HM_ID'
        }
        cp_data = {
            'LoadBalancingPolicy': {
                'version': '1.0',
                'data': policy_data
            }
        }
        cp.data = cp_data
        self.lb_driver.member_add.side_effect = ['MEMBER1_ID', 'MEMBER2_ID']
        m_node_get.side_effect = [node1, node2]
        m_extract.return_value = policy_data
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver

        # do it
        candidates = ['NODE1_ID', 'NODE2_ID']
        res = policy._add_member(action.context, candidates,
                                 cp, self.lb_driver)

        # assertions
        self.assertEqual([], res)
        m_extract.assert_called_once_with(cp_data)
        calls_node_get = [
            mock.call('action_context', node_id='NODE1_ID'),
            mock.call('action_context', node_id='NODE2_ID')
        ]
        m_node_get.assert_has_calls(calls_node_get)
        calls_node_update = [
            mock.call(action.context, 'NODE1_ID', mock.ANY),
            mock.call(action.context, 'NODE2_ID', mock.ANY)
        ]
        m_node_update.assert_has_calls(calls_node_update)
        calls_member_add = [
            mock.call(node1, 'LB_ID', 'POOL_ID', 80, 'test-subnet'),
            mock.call(node2, 'LB_ID', 'POOL_ID', 80, 'test-subnet'),
        ]
        self.lb_driver.member_add.assert_has_calls(calls_member_add)

    @mock.patch.object(no.Node, 'get')
    @mock.patch.object(no.Node, 'update')
    def test__add_member_fail(self, m_node_update, m_node_get,
                              m_extract, m_load):
        node1 = mock.Mock(id='NODE1_ID', data={})
        action = mock.Mock(context='action_context',
                           action=consts.CLUSTER_RESIZE,
                           data={
                               'creation': {
                                   'nodes': ['NODE1_ID']
                               }
                           })
        cp = mock.Mock()
        policy_data = {
            'loadbalancer': 'LB_ID',
            'listener': 'LISTENER_ID',
            'pool': 'POOL_ID',
            'healthmonitor': 'HM_ID'
        }
        cp_data = {
            'LoadBalancingPolicy': {
                'version': '1.0',
                'data': policy_data
            }
        }
        cp.data = cp_data
        self.lb_driver.member_add.return_value = None
        m_node_get.return_value = node1
        m_extract.return_value = policy_data
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver

        # do it
        candidates = ['NODE1_ID']
        res = policy._add_member(action.context, candidates,
                                 cp, self.lb_driver)

        # assertions
        self.assertEqual(['NODE1_ID'], res)
        m_extract.assert_called_once_with(cp_data)
        m_node_get.assert_called_once_with(
            'action_context', node_id='NODE1_ID')
        m_node_update.assert_called_once_with(
            'action_context', 'NODE1_ID', mock.ANY)
        self.lb_driver.member_add.assert_called_once_with(
            node1, 'LB_ID', 'POOL_ID', 80, 'test-subnet')

    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_add_member')
    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_remove_member')
    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_get_post_candidates')
    def test_post_op_node_create(self, m_get, m_remove, m_add,
                                 m_candidates, m_load):
        ctx = mock.Mock()
        cid = 'CLUSTER_ID'
        cluster = mock.Mock(user='user1', project='project1')
        action = mock.Mock(data={}, context=ctx, action=consts.NODE_CREATE,
                           node=mock.Mock(id='NODE_ID'))
        action.entity = cluster
        cp = mock.Mock()
        m_load.return_value = cp
        m_add.return_value = []
        m_get.return_value = ['NODE_ID']

        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver
        # do it
        res = policy.post_op(cid, action)

        # assertion
        self.assertIsNone(res)
        m_get.assert_called_once_with(action)
        m_load.assert_called_once_with(ctx, cid, policy.id)
        m_add.assert_called_once_with(ctx, ['NODE_ID'], cp, self.lb_driver)
        self.assertFalse(m_remove.called)

    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_add_member')
    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_remove_member')
    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_get_post_candidates')
    def test_post_op_add_nodes(self, m_get, m_remove, m_add,
                               m_candidates, m_load):
        cid = 'CLUSTER_ID'
        cluster = mock.Mock(user='user1', project='project1')
        action = mock.Mock(context='action_context',
                           action=consts.CLUSTER_RESIZE,
                           data={
                               'creation': {
                                   'nodes': ['NODE1_ID', 'NODE2_ID']
                               }
                           })
        action.entity = cluster
        candidates = ['NODE1_ID', 'NODE2_ID']
        m_get.return_value = candidates
        cp = mock.Mock()
        m_load.return_value = cp
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver

        # do it
        res = policy.post_op(cid, action)

        # assertions
        self.assertIsNone(res)
        m_get.assert_called_once_with(action)
        m_load.assert_called_once_with('action_context', cid, policy.id)
        m_add.assert_called_once_with(action.context, candidates,
                                      cp, self.lb_driver)
        self.assertFalse(m_remove.called)

    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_add_member')
    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_process_recovery')
    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_get_post_candidates')
    def test_post_op_node_recover(self, m_get, m_recovery, m_add,
                                  m_candidates, m_load):
        cid = 'CLUSTER_ID'
        node = mock.Mock(user='user1', project='project1', id='NODE1')
        action = mock.Mock(context='action_context',
                           action=consts.NODE_RECOVER,
                           data={},
                           outputs={})
        action.entity = node
        m_recovery.return_value = ['NODE1']
        m_get.return_value = ['NODE1']
        cp = mock.Mock()
        m_load.return_value = cp
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver

        # do it
        res = policy.post_op(cid, action)

        # assertions
        self.assertIsNone(res)
        m_get.assert_called_once_with(action)
        m_load.assert_called_once_with('action_context', cid, policy.id)
        m_add.assert_called_once_with(action.context, ['NODE1'],
                                      cp, self.lb_driver)
        m_recovery.assert_called_once_with(['NODE1'], cp, self.lb_driver,
                                           action)

    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_add_member')
    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_remove_member')
    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_get_post_candidates')
    def test_post_op_clusterresize_failed(self, m_get, m_remove, m_add,
                                          m_candidates, m_load):
        cluster_id = 'CLUSTER_ID'
        action = mock.Mock(data={'creation': {'nodes': ['NODE1_ID']}},
                           context='action_context',
                           action=consts.CLUSTER_RESIZE)

        cp = mock.Mock()
        m_load.return_value = cp
        m_get.return_value = ['NODE1_ID']
        m_add.return_value = ['NODE1_ID']
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver

        res = policy.post_op(cluster_id, action)

        self.assertIsNone(res)
        self.assertEqual(policy_base.CHECK_ERROR, action.data['status'])
        self.assertEqual("Failed in adding nodes into lb pool: "
                         "['NODE1_ID']", action.data['reason'])
        m_get.assert_called_once_with(action)
        m_add.assert_called_once_with(action.context, ['NODE1_ID'],
                                      cp, self.lb_driver)
        self.assertFalse(m_remove.called)

    @mock.patch.object(no.Node, 'get')
    @mock.patch.object(no.Node, 'update')
    def test__remove_member(self, m_node_update, m_node_get,
                            m_extract, m_load):
        node1 = mock.Mock(id='NODE1', data={'lb_member': 'MEM_ID1'})
        node2 = mock.Mock(id='NODE2', data={'lb_member': 'MEM_ID2'})
        action = mock.Mock(
            context='action_context', action=consts.CLUSTER_DEL_NODES,
            data={
                'deletion': {
                    'count': 2,
                    'candidates': ['NODE1', 'NODE2']
                }
            })
        cp = mock.Mock()
        policy_data = {
            'loadbalancer': 'LB_ID',
            'listener': 'LISTENER_ID',
            'pool': 'POOL_ID',
            'healthmonitor': 'HM_ID'
        }
        cp_data = {
            'LoadBalancingPolicy': {
                'version': '1.0', 'data': policy_data
            }
        }
        cp.data = cp_data
        self.lb_driver.member_remove.return_value = True
        m_node_get.side_effect = [node1, node2]
        m_extract.return_value = policy_data
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver

        candidates = [node1.id, node2.id]
        res = policy._remove_member(action.context, candidates,
                                    cp, self.lb_driver)

        m_extract.assert_called_once_with(cp_data)
        calls_node_get = [
            mock.call(action.context, node_id='NODE1'),
            mock.call(action.context, node_id='NODE2')
        ]
        m_node_get.assert_has_calls(calls_node_get)
        calls_node_update = [
            mock.call(action.context, 'NODE1', mock.ANY),
            mock.call(action.context, 'NODE2', mock.ANY)
        ]
        m_node_update.assert_has_calls(calls_node_update)
        calls_member_del = [
            mock.call('LB_ID', 'POOL_ID', 'MEM_ID1'),
            mock.call('LB_ID', 'POOL_ID', 'MEM_ID2')
        ]
        self.lb_driver.member_remove.assert_has_calls(calls_member_del)
        self.assertEqual([], res)

    @mock.patch.object(no.Node, 'get')
    @mock.patch.object(no.Node, 'update')
    def test__remove_member_not_in_pool(self, m_node_update, m_node_get,
                                        m_extract, m_load):
        node1 = mock.Mock(id='NODE1', data={'lb_member': 'MEM_ID1'})
        node2 = mock.Mock(id='NODE2', data={})
        action = mock.Mock(
            context='action_context', action=consts.CLUSTER_DEL_NODES,
            data={
                'deletion': {
                    'count': 2,
                    'candidates': ['NODE1', 'NODE2']
                }
            })
        cp = mock.Mock()
        policy_data = {
            'loadbalancer': 'LB_ID',
            'listener': 'LISTENER_ID',
            'pool': 'POOL_ID',
            'healthmonitor': 'HM_ID'
        }
        cp_data = {
            'LoadBalancingPolicy': {
                'version': '1.0', 'data': policy_data
            }
        }
        cp.data = cp_data
        self.lb_driver.member_remove.return_value = True
        m_node_get.side_effect = [node1, node2]
        m_extract.return_value = policy_data
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver

        candidates = [node1.id, node2.id]
        res = policy._remove_member(action.context, candidates,
                                    cp, self.lb_driver)

        m_extract.assert_called_once_with(cp_data)
        calls_node_get = [
            mock.call(action.context, node_id='NODE1'),
            mock.call(action.context, node_id='NODE2')
        ]
        m_node_get.assert_has_calls(calls_node_get)
        m_node_update.assert_called_once_with(
            action.context, 'NODE1', mock.ANY)
        self.lb_driver.member_remove.assert_called_once_with(
            'LB_ID', 'POOL_ID', 'MEM_ID1')
        self.assertEqual([], res)

    @mock.patch.object(no.Node, 'get')
    @mock.patch.object(no.Node, 'update')
    def test__remove_member_fail(self, m_node_update, m_node_get,
                                 m_extract, m_load):
        node1 = mock.Mock(id='NODE1', data={'lb_member': 'MEM_ID1'})
        action = mock.Mock(
            context='action_context', action=consts.CLUSTER_DEL_NODES,
            data={
                'deletion': {
                    'count': 1,
                    'candidates': ['NODE1']
                }
            })
        cp = mock.Mock()
        policy_data = {
            'loadbalancer': 'LB_ID',
            'listener': 'LISTENER_ID',
            'pool': 'POOL_ID',
            'healthmonitor': 'HM_ID'
        }
        cp_data = {
            'LoadBalancingPolicy': {
                'version': '1.0', 'data': policy_data
            }
        }
        cp.data = cp_data
        self.lb_driver.member_remove.return_value = False
        m_node_get.return_value = node1
        m_extract.return_value = policy_data
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver

        candidates = [node1.id]
        res = policy._remove_member(action.context, candidates,
                                    cp, self.lb_driver)

        m_extract.assert_called_once_with(cp_data)
        m_node_get.assert_called_once_with(action.context, node_id='NODE1')
        m_node_update.assert_called_once_with(
            action.context, 'NODE1', mock.ANY)
        self.lb_driver.member_remove.assert_called_once_with(
            'LB_ID', 'POOL_ID', 'MEM_ID1')
        self.assertEqual(['NODE1'], res)

    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_remove_member')
    def test_pre_op_del_nodes_ok(self, m_remove, m_candidates, m_load):
        cluster_id = 'CLUSTER_ID'
        cluster = mock.Mock(user='user1', project='project1')
        action = mock.Mock(
            context='action_context', action=consts.CLUSTER_DEL_NODES,
            data={
                'deletion': {
                    'count': 2,
                    'candidates': ['NODE1_ID', 'NODE2_ID']
                }
            })
        action.entity = cluster
        m_candidates.return_value = ['NODE1_ID', 'NODE2_ID']
        cp = mock.Mock()
        m_load.return_value = cp
        m_remove.return_value = []
        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver

        res = policy.pre_op(cluster_id, action)
        self.assertIsNone(res)

        m_load.assert_called_once_with('action_context', cluster_id, policy.id)

        expected_data = {'deletion': {'candidates': ['NODE1_ID', 'NODE2_ID'],
                                      'count': 2}}
        self.assertEqual(expected_data, action.data)

    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_remove_member')
    def test_pre_op_del_nodes_failed(self, m_remove, m_candidates, m_load):
        cluster_id = 'CLUSTER_ID'
        cluster = mock.Mock(user='user1', project='project1')
        action = mock.Mock(
            action=consts.CLUSTER_RESIZE,
            context='action_context',
            data={'deletion': {'candidates': ['NODE1_ID']}})
        action.entity = cluster
        m_remove.return_value = ['NODE1_ID']

        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        policy._lbaasclient = self.lb_driver
        res = policy.pre_op(cluster_id, action)

        self.assertIsNone(res)
        self.assertEqual(policy_base.CHECK_ERROR, action.data['status'])
        self.assertEqual("Failed in removing deleted node(s) from lb pool: "
                         "['NODE1_ID']", action.data['reason'])

        m_remove.assert_called_once_with(action.context, ['NODE1_ID'],
                                         mock.ANY, self.lb_driver)

    @mock.patch.object(no.Node, 'update')
    def test__process_recovery_not_lb_member(self, m_update, m1, m2):
        node = mock.Mock(id='NODE', data={})
        action = mock.Mock(
            action=consts.NODE_RECOVER,
            context='action_context')
        action.entity = node

        cp = mock.Mock()

        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        res = policy._process_recovery(['NODE'], cp, self.lb_driver, action)

        self.assertEqual(['NODE'], res)
        m_update.assert_called_once_with(action.context, {'data': {}})

    @mock.patch.object(no.Node, 'update')
    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_remove_member')
    def test__process_recovery_reboot(self, m_remove, m_update, m1, m2):
        node = mock.Mock(id='NODE', data={'lb_member': 'mem_1'})
        action = mock.Mock(
            action=consts.NODE_RECOVER,
            context='action_context')
        action.entity = node

        cp = mock.Mock()

        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        res = policy._process_recovery(['NODE'], cp, self.lb_driver, action)

        self.assertIsNone(res)

        self.assertFalse(m_remove.called)
        self.assertFalse(m_update.called)

    @mock.patch.object(no.Node, 'update')
    @mock.patch.object(lb_policy.LoadBalancingPolicy, '_remove_member')
    def test__process_recovery_recreate(self, m_remove, m_update, m1, m2):
        node = mock.Mock(id='NODE', data={'lb_member': 'mem_1',
                                          'recovery': 'RECREATE'})
        action = mock.Mock(
            action=consts.NODE_RECOVER,
            context='action_context')
        action.entity = node

        cp = mock.Mock()

        policy = lb_policy.LoadBalancingPolicy('test-policy', self.spec)
        res = policy._process_recovery(['NODE'], cp, self.lb_driver, action)

        self.assertEqual(['NODE'], res)
        m_remove.assert_called_once_with(action.context, ['NODE'], cp,
                                         self.lb_driver, handle_err=False)
        m_update.assert_called_once_with(action.context, {'data': {}})



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections

import mock
import six

from senlin.common import constraints
from senlin.common import exception as exc
from senlin.common import schema
from senlin.tests.unit.common import base


class FakeSchema(schema.SchemaBase):

    def __getitem__(self, key):
        if key == self.TYPE:
            return self.STRING
        return super(FakeSchema, self).__getitem__(key)

    def resolve(self, value):
        return str(value)

    def validate(self, value, context=None):
        return


class TestAnyIndexDict(base.SenlinTestCase):

    def test_basic(self):
        sot = schema.AnyIndexDict('*')

        self.assertIsInstance(sot, collections.Mapping)

        self.assertEqual('*', sot.value)
        self.assertEqual('*', sot[1])
        self.assertEqual('*', sot[2])
        self.assertEqual('*', sot['*'])

        for a in sot:
            self.assertEqual('*', a)

        self.assertEqual(1, len(sot))

    def test_bad_index(self):
        sot = schema.AnyIndexDict('*')

        ex = self.assertRaises(KeyError, sot.__getitem__, 'foo')

        # the following test is not interesting
        self.assertEqual("'Invalid key foo'", str(ex))


class TestSchemaBase(base.SenlinTestCase):

    def test_basic(self):
        sot = FakeSchema(description='desc', default='default', required=True,
                         schema=None, constraints=None, min_version='1.0',
                         max_version='2.0')
        self.assertEqual('desc', sot.description)
        self.assertEqual('default', sot.default)
        self.assertTrue(sot.required)
        self.assertIsNone(sot.schema)
        self.assertEqual([], sot.constraints)
        self.assertEqual('1.0', sot.min_version)
        self.assertEqual('2.0', sot.max_version)
        self.assertTrue(sot.has_default())

    def test_init_schema_invalid(self):
        ex = self.assertRaises(exc.ESchema, FakeSchema, schema=mock.Mock())
        self.assertEqual('Schema valid only for List or Map, not String',
                         six.text_type(ex))

    def test_get_default(self):
        sot = FakeSchema(default='DEFAULT')
        mock_resolve = self.patchobject(sot, 'resolve', return_value='VVV')

        res = sot.get_default()

        self.assertEqual('VVV', res)
        mock_resolve.assert_called_once_with('DEFAULT')

    def test__validate_default(self):
        sot = FakeSchema()

        self.assertIsNone(sot._validate_default(mock.Mock()))

    def test__validate_default_with_value(self):
        sot = FakeSchema(default='DEFAULT')
        mock_validate = self.patchobject(sot, 'validate', return_value=None)
        fake_context = mock.Mock()

        res = sot._validate_default(fake_context)

        self.assertIsNone(res)
        mock_validate.assert_called_once_with('DEFAULT', fake_context)

    def test__validate_default_with_value_but_failed(self):
        sot = FakeSchema(default='DEFAULT')
        mock_validate = self.patchobject(sot, 'validate',
                                         side_effect=ValueError('boom'))
        fake_context = mock.Mock()

        ex = self.assertRaises(exc.ESchema,
                               sot._validate_default,
                               fake_context)

        mock_validate.assert_called_once_with('DEFAULT', fake_context)
        self.assertEqual('Invalid default DEFAULT: boom', six.text_type(ex))

    def test_validate_constraints(self):
        c1 = mock.Mock()
        c2 = mock.Mock()
        sot = FakeSchema(constraints=[c1, c2])
        ctx = mock.Mock()

        res = sot.validate_constraints('VALUE', context=ctx)

        self.assertIsNone(res)
        c1.validate.assert_called_once_with('VALUE', schema=None, context=ctx)
        c2.validate.assert_called_once_with('VALUE', schema=None, context=ctx)

    def test_validate_constraints_failed(self):
        c1 = mock.Mock()
        c1.validate.side_effect = ValueError('BOOM')
        sot = FakeSchema(constraints=[c1])
        ctx = mock.Mock()

        ex = self.assertRaises(exc.ESchema,
                               sot.validate_constraints,
                               'FOO', context=ctx)

        c1.validate.assert_called_once_with('FOO', schema=None, context=ctx)
        self.assertEqual('BOOM', six.text_type(ex))

    def test__validate_version(self):
        sot = FakeSchema(min_version='1.0', max_version='2.0')

        res = sot._validate_version('field', '1.0')
        self.assertIsNone(res)

        res = sot._validate_version('field', '1.1')
        self.assertIsNone(res)

        # there is a warning, but validation passes
        res = sot._validate_version('field', '2.0')
        self.assertIsNone(res)

        ex = self.assertRaises(exc.ESchema,
                               sot._validate_version,
                               'field', '0.9')
        self.assertEqual('field (min_version=1.0) is not supported by '
                         'spec version 0.9.',
                         six.text_type(ex))

        ex = self.assertRaises(exc.ESchema,
                               sot._validate_version,
                               'field', '2.1')
        self.assertEqual('field (max_version=2.0) is not supported by '
                         'spec version 2.1.',
                         six.text_type(ex))

    def test__validate_version_no_min_version(self):
        sot = FakeSchema(max_version='2.0')

        res = sot._validate_version('field', '1.0')
        self.assertIsNone(res)

        res = sot._validate_version('field', '2.0')
        self.assertIsNone(res)

        ex = self.assertRaises(exc.ESchema,
                               sot._validate_version,
                               'field', '2.1')
        self.assertEqual('field (max_version=2.0) is not supported by '
                         'spec version 2.1.',
                         six.text_type(ex))

    def test__validate_version_no_max_version(self):
        sot = FakeSchema(min_version='1.0')

        res = sot._validate_version('field', '1.0')
        self.assertIsNone(res)

        res = sot._validate_version('field', '2.3')
        self.assertIsNone(res)

        ex = self.assertRaises(exc.ESchema,
                               sot._validate_version,
                               'field', '0.5')
        self.assertEqual('field (min_version=1.0) is not supported by '
                         'spec version 0.5.',
                         six.text_type(ex))

    def test__validate_version_no_version_restriction(self):
        sot = FakeSchema()

        res = sot._validate_version('field', '1.0')
        self.assertIsNone(res)

        res = sot._validate_version('field', '2.3')
        self.assertIsNone(res)

    def test__getitem__(self):
        sot = FakeSchema(description='desc', default='default', required=False,
                         constraints=[{'foo': 'bar'}])

        self.assertEqual('desc', sot['description'])
        self.assertEqual('default', sot['default'])
        self.assertEqual(False, sot['required'])
        self.assertEqual([{'foo': 'bar'}], sot['constraints'])
        self.assertRaises(KeyError, sot.__getitem__, 'bogus')

        sot = schema.List(schema=schema.String())
        self.assertEqual(
            {
                '*': {
                    'required': False,
                    'type': 'String',
                    'updatable': False
                }
            },
            sot['schema'])

    def test__iter__(self):
        sot = FakeSchema(description='desc', default='default', required=False,
                         constraints=[{'foo': 'bar'}])

        res = list(iter(sot))

        self.assertIn('type', res)
        self.assertIn('description', res)
        self.assertIn('default', res)
        self.assertIn('required', res)
        self.assertIn('constraints', res)

    def test__len__(self):
        sot = FakeSchema()

        res = list(iter(sot))

        self.assertIn('type', res)
        self.assertIn('required', res)
        self.assertEqual(2, len(sot))


class TestPropertySchema(base.SenlinTestCase):

    def setUp(self):
        super(TestPropertySchema, self).setUp()

        class TestProperty(schema.PropertySchema):

            def __getitem__(self, key):
                if key == self.TYPE:
                    return 'TEST'
                return super(TestProperty, self).__getitem__(key)

        self.cls = TestProperty

    def test_basic(self):
        sot = self.cls()

        self.assertIsNone(sot.description)
        self.assertIsNone(sot.default)
        self.assertFalse(sot.required)
        self.assertIsNone(sot.schema)
        self.assertEqual([], sot.constraints)
        self.assertIsNone(sot.min_version)
        self.assertIsNone(sot.max_version)
        self.assertFalse(sot.updatable)

    def test__getitem__(self):
        sot = self.cls(updatable=True)

        res = sot['updatable']

        self.assertTrue(res)
        self.assertTrue(sot.updatable)


class TestBoolean(base.SenlinTestCase):

    def test_basic(self):
        sot = schema.Boolean('desc')

        self.assertEqual('Boolean', sot['type'])
        self.assertEqual('desc', sot['description'])

    def test_to_schema_type(self):
        sot = schema.Boolean('desc')

        res = sot.to_schema_type(True)
        self.assertTrue(res)

        res = sot.to_schema_type('true')
        self.assertTrue(res)

        res = sot.to_schema_type('trUE')
        self.assertTrue(res)

        res = sot.to_schema_type('False')
        self.assertFalse(res)

        res = sot.to_schema_type('FALSE')
        self.assertFalse(res)

        ex = self.assertRaises(exc.ESchema, sot.to_schema_type, 'bogus')
        self.assertEqual("The value 'bogus' is not a valid Boolean",
                         six.text_type(ex))

    def test_resolve(self):
        sot = schema.Boolean()

        res = sot.resolve(True)
        self.assertTrue(res)

        res = sot.resolve(False)
        self.assertFalse(res)

        res = sot.resolve('Yes')
        self.assertTrue(res)

    def test_validate(self):
        sot = schema.Boolean()

        res = sot.validate(True)
        self.assertIsNone(res)

        res = sot.validate('No')
        self.assertIsNone(res)

        ex = self.assertRaises(exc.ESchema, sot.validate, 'bogus')
        self.assertEqual("The value 'bogus' is not a valid Boolean",
                         six.text_type(ex))


class TestInteger(base.SenlinTestCase):

    def test_basic(self):
        sot = schema.Integer('desc')

        self.assertEqual('Integer', sot['type'])
        self.assertEqual('desc', sot['description'])

    def test_to_schema_type(self):
        sot = schema.Integer('desc')

        res = sot.to_schema_type(123)
        self.assertEqual(123, res)

        res = sot.to_schema_type('123')
        self.assertEqual(123, res)

        res = sot.to_schema_type(False)
        self.assertEqual(0, res)

        self.assertIsNone(sot.to_schema_type(None))

        ex = self.assertRaises(exc.ESchema, sot.to_schema_type, '456L')
        self.assertEqual("The value '456L' is not a valid Integer",
                         six.text_type(ex))

    def test_resolve(self):
        sot = schema.Integer()

        res = sot.resolve(1)
        self.assertEqual(1, res)

        res = sot.resolve(True)
        self.assertEqual(1, res)

        res = sot.resolve(False)
        self.assertEqual(0, res)

        self.assertIsNone(sot.resolve(None))

        ex = self.assertRaises(exc.ESchema, sot.resolve, '456L')
        self.assertEqual("The value '456L' is not a valid Integer",
                         six.text_type(ex))

    def test_validate(self):
        sot = schema.Integer()

        res = sot.validate(1)
        self.assertIsNone(res)

        res = sot.validate('1')
        self.assertIsNone(res)

        res = sot.validate(True)
        self.assertIsNone(res)

        mock_constraints = self.patchobject(sot, 'validate_constraints',
                                            return_value=None)

        res = sot.validate(1)
        self.assertIsNone(res)
        mock_constraints.assert_called_once_with(1, schema=sot, context=None)
        ex = self.assertRaises(exc.ESchema, sot.validate, 'bogus')
        self.assertEqual("The value 'bogus' is not a valid Integer",
                         six.text_type(ex))


class TestString(base.SenlinTestCase):

    def test_basic(self):
        sot = schema.String('desc')

        self.assertEqual('String', sot['type'])
        self.assertEqual('desc', sot['description'])

    def test_invalid_constructor(self):
        self.assertRaises(exc.ESchema,
                          schema.String,
                          schema=schema.String('String'))

    def test_to_schema_type(self):
        sot = schema.String('desc')

        res = sot.to_schema_type(123)
        self.assertEqual('123', res)

        res = sot.to_schema_type('123')
        self.assertEqual('123', res)

        res = sot.to_schema_type(False)
        self.assertEqual('False', res)

        res = sot.to_schema_type(None)
        self.assertIsNone(res)

        res = sot.to_schema_type(u'\u4e2d\u6587')
        self.assertEqual(u'\u4e2d\u6587', res)

    def test_resolve(self):
        sot = schema.String()

        res = sot.resolve(1)
        self.assertEqual('1', res)

        res = sot.resolve(True)
        self.assertEqual('True', res)

        res = sot.resolve(None)
        self.assertIsNone(res)

    def test_validate(self):
        sot = schema.String()

        res = sot.validate('1')
        self.assertIsNone(res)

        res = sot.validate(u'unicode')
        self.assertIsNone(res)

        ex = self.assertRaises(exc.ESchema, sot.validate, 1)
        self.assertEqual("The value '1' is not a valid string.",
                         six.text_type(ex))

        mock_constraints = self.patchobject(sot, 'validate_constraints',
                                            return_value=None)

        res = sot.validate("abcd")
        self.assertIsNone(res)
        mock_constraints.assert_called_once_with(
            "abcd", schema=sot, context=None)


class TestNumber(base.SenlinTestCase):

    def test_basic(self):
        sot = schema.Number('desc')

        self.assertEqual('Number', sot['type'])
        self.assertEqual('desc', sot['description'])

    def test_to_schema_type(self):
        sot = schema.Number('desc')

        res = sot.to_schema_type(123)
        self.assertEqual(123, res)

        res = sot.to_schema_type(123.34)
        self.assertEqual(123.34, res)

        res = sot.to_schema_type(False)
        self.assertEqual(False, res)

    def test_resolve(self):
        sot = schema.Number()
        mock_convert = self.patchobject(sot, 'to_schema_type')

        res = sot.resolve(1)
        self.assertEqual(mock_convert.return_value, res)
        mock_convert.assert_called_once_with(1)

    def test_validate(self):
        sot = schema.Number()

        res = sot.validate(1)
        self.assertIsNone(res)

        res = sot.validate('1')
        self.assertIsNone(res)

        ex = self.assertRaises(exc.ESchema, sot.validate, "bogus")
        self.assertEqual("The value 'bogus' is not a valid number.",
                         six.text_type(ex))

        mock_constraints = self.patchobject(sot, 'validate_constraints',
                                            return_value=None)

        res = sot.validate('1234')
        self.assertIsNone(res)
        mock_constraints.assert_called_once_with(
            1234, schema=sot, context=None)


class TestList(base.SenlinTestCase):

    def test_basic(self):
        sot = schema.List('desc')

        self.assertEqual('List', sot['type'])
        self.assertEqual('desc', sot['description'])

    def test__get_children(self):
        sot = schema.List('desc', schema=schema.String())

        res = sot._get_children(['v1', 'v2'], [0, 1])
        self.assertEqual(['v1', 'v2'], list(res))

    def test_resolve(self):
        sot = schema.List(schema=schema.String())

        res = sot.resolve(['v1', 'v2'])

        self.assertEqual(['v1', 'v2'], res)

        self.assertRaises(TypeError,
                          sot.resolve,
                          123)

    def test_validate(self):
        sot = schema.List(schema=schema.String())

        res = sot.validate(['abc', 'def'])

        self.assertIsNone(res)

    def test_validate_failed(self):
        sot = schema.List(schema=schema.String())

        ex = self.assertRaises(exc.ESchema, sot.validate, None)
        self.assertEqual("'None' is not a List", six.text_type(ex))


class TestMap(base.SenlinTestCase):

    def test_basic(self):
        sot = schema.Map('desc')

        self.assertEqual('Map', sot['type'])
        self.assertEqual('desc', sot['description'])

    def test__get_children(self):
        sot = schema.Map('desc', schema={'foo': schema.String()})

        res = sot._get_children({'foo': 'bar'})

        self.assertEqual({'foo': 'bar'}, dict(res))

    def test_get_default(self):
        sot = schema.Map(schema={'foo': schema.String()})
        self.assertEqual({}, sot.get_default())

        sot = schema.Map(default={'foo': 'bar'},
                         schema={'foo': schema.String()})
        self.assertEqual({'foo': 'bar'}, sot.get_default())

        sot = schema.Map(default='bad', schema={'foo': schema.String()})
        ex = self.assertRaises(exc.ESchema, sot.get_default)
        self.assertEqual("'bad' is not a Map", six.text_type(ex))

    def test_resolve(self):
        sot = schema.Map(schema={'foo': schema.String()})

        res = sot.resolve({"foo": "bar"})
        self.assertEqual({'foo': 'bar'}, res)

        res = sot.resolve('{"foo": "bar"}')
        self.assertEqual({'foo': 'bar'}, res)

        ex = self.assertRaises(exc.ESchema, sot.resolve, 'plainstring')
        self.assertEqual("'plainstring' is not a Map", six.text_type(ex))

    def test_validate(self):
        sot = schema.Map(schema={'foo': schema.String()})

        res = sot.validate({"foo": "bar"})

        self.assertIsNone(res)

    def test_validate_failed(self):
        sot = schema.Map(schema={'foo': schema.String()})

        ex = self.assertRaises(exc.ESchema, sot.validate, None)
        self.assertEqual("'None' is not a Map", six.text_type(ex))

        ex = self.assertRaises(exc.ESchema, sot.validate, 'bogus')
        self.assertEqual("'bogus' is not a Map", six.text_type(ex))


class TestStringParam(base.SenlinTestCase):

    def test_basic(self):
        sot = schema.StringParam()
        self.assertEqual('String', sot['type'])
        self.assertEqual(False, sot['required'])

    def test_validate(self):
        sot = schema.StringParam()
        result = sot.validate('foo')
        self.assertIsNone(result)

    def test_validate_bad_type(self):
        sot = schema.StringParam()
        self.assertRaises(TypeError,
                          sot.validate,
                          ['123'])

    def test_validate_failed_constraint(self):
        sot = schema.StringParam(
            constraints=[constraints.AllowedValues(('abc', 'def'))])

        ex = self.assertRaises(exc.ESchema, sot.validate, '123')

        self.assertEqual("'123' must be one of the allowed values: abc, def",
                         six.text_type(ex))


class TestIntegerParam(base.SenlinTestCase):

    def test_basic(self):
        sot = schema.IntegerParam()
        self.assertEqual('Integer', sot['type'])
        self.assertEqual(False, sot['required'])

    def test_validate(self):
        sot = schema.IntegerParam()
        result = sot.validate(123)
        self.assertIsNone(result)

    def test_validate_bad_type(self):
        sot = schema.IntegerParam()
        self.assertRaises(ValueError,
                          sot.validate,
                          'not int')

    def test_validate_failed_constraint(self):
        sot = schema.IntegerParam(
            constraints=[constraints.AllowedValues((123, 124))])

        ex = self.assertRaises(exc.ESchema, sot.validate, 12)

        self.assertEqual("'12' must be one of the allowed values: 123, 124",
                         six.text_type(ex))


class TestOperation(base.SenlinTestCase):

    def test_basic(self):
        sot = schema.Operation()
        self.assertEqual('Undocumented', sot['description'])
        self.assertEqual({}, sot['parameters'])

    def test_initialized(self):
        sot = schema.Operation('des', schema={'foo': schema.StringParam()})
        self.assertEqual('des', sot['description'])
        self.assertEqual({'foo': {'required': False, 'type': 'String'}},
                         sot['parameters'])

    def test_validate(self):
        sot = schema.Operation('des', schema={'foo': schema.StringParam()})
        res = sot.validate({'foo': 'bar'})
        self.assertIsNone(res)

    def test_validate_unrecognizable_param(self):
        sot = schema.Operation('des', schema={'foo': schema.StringParam()})

        ex = self.assertRaises(exc.ESchema, sot.validate,
                               {'baar': 'baar'})

        self.assertEqual("Unrecognizable parameter 'baar'", six.text_type(ex))

    def test_validate_failed_type(self):
        sot = schema.Operation('des', schema={'foo': schema.StringParam()})

        ex = self.assertRaises(exc.ESchema, sot.validate,
                               {'foo': ['baaar']})

        self.assertEqual("value is not a string",
                         six.text_type(ex))

    def test_validate_failed_constraint(self):
        sot = schema.Operation(
            'des',
            schema={
                'foo': schema.StringParam(
                    constraints=[constraints.AllowedValues(['bar'])])
            }
        )

        ex = self.assertRaises(exc.ESchema, sot.validate,
                               {'foo': 'baaar'})

        self.assertEqual("'baaar' must be one of the allowed values: bar",
                         six.text_type(ex))

    def test_validate_failed_required(self):
        sot = schema.Operation(
            'des',
            schema={
                'foo': schema.StringParam(),
                'bar': schema.StringParam(required=True)
            }
        )

        ex = self.assertRaises(exc.ESchema, sot.validate,
                               {'foo': 'baaar'})

        self.assertEqual("Required parameter 'bar' not provided",
                         six.text_type(ex))

    def test_validate_failed_version(self):
        sot = schema.Operation(
            'des',
            schema={
                'foo': schema.StringParam(min_version='2.0'),
            }
        )

        ex = self.assertRaises(exc.ESchema, sot.validate,
                               {'foo': 'baaar'}, '1.0')

        self.assertEqual("foo (min_version=2.0) is not supported by spec "
                         "version 1.0.", six.text_type(ex))


class TestSpec(base.SenlinTestCase):
    spec_schema = {
        'key1': schema.String('first key', default='value1'),
        'key2': schema.Integer('second key', required=True),
    }

    def test_init(self):
        data = {'key1': 'value1', 'key2': 2}
        sot = schema.Spec(self.spec_schema, data)

        self.assertEqual(self.spec_schema, sot._schema)
        self.assertEqual(data, sot._data)
        self.assertIsNone(sot._version)

    def test_init_with_version(self):
        data = {'key1': 'value1', 'key2': 2}
        sot = schema.Spec(self.spec_schema, data, version='1.2')

        self.assertEqual(self.spec_schema, sot._schema)
        self.assertEqual(data, sot._data)
        self.assertEqual('1.2', sot._version)

    def test_validate(self):
        data = {'key1': 'value1', 'key2': 2}
        sot = schema.Spec(self.spec_schema, data)
        res = sot.validate()
        self.assertIsNone(res)

        data1 = {'key2': 2}
        sot = schema.Spec(self.spec_schema, data1)
        res = sot.validate()
        self.assertIsNone(res)

    def test_validate_fail_unrecognizable_key(self):
        spec_schema = {
            'key1': schema.String('first key', default='value1'),
        }
        data = {'key1': 'value1', 'key2': 2}
        sot = schema.Spec(spec_schema, data, version='1.0')
        ex = self.assertRaises(exc.ESchema, sot.validate)

        self.assertIn("Unrecognizable spec item 'key2'",
                      six.text_type(ex.message))

    def test_validate_fail_value_type_incorrect(self):
        spec_schema = {
            'key1': schema.String('first key', default='value1'),
            'key2': schema.Integer('second key', required=True),
        }

        data = {'key1': 'value1', 'key2': 'abc'}
        spec = schema.Spec(spec_schema, data, version='1.0')
        ex = self.assertRaises(exc.ESchema, spec.validate)
        self.assertIn("The value 'abc' is not a valid Integer",
                      six.text_type(ex.message))

    def test_validate_version_good(self):
        spec_schema = {
            'type': schema.String('Type name', required=True),
            'version': schema.String('Version number', required=True),
            'key1': schema.String('first key', default='value1'),
            'key2': schema.Integer('second key', required=True,
                                   min_version='1.0', max_version='1.2'),
        }

        data = {
            'key1': 'value1',
            'key2': 2,
            'type': 'test-type',
            'version': '1.0'
        }
        spec = schema.Spec(spec_schema, data)
        self.assertIsNone(spec.validate())

        data = {'key2': 2, 'type': 'test-type', 'version': '1.2'}
        spec = schema.Spec(spec_schema, data)
        self.assertIsNone(spec.validate())

    def test_validate_version_fail_unsupported_version(self):
        spec_schema = {
            'type': schema.String('Type name', required=True),
            'version': schema.String('Version number', required=True),
            'key1': schema.String('first key', default='value1',
                                  min_version='1.1'),
            'key2': schema.Integer('second key', required=True),
        }

        data = {
            'key1': 'value1',
            'key2': 2,
            'type': 'test-type',
            'version': '1.0'
        }
        spec = schema.Spec(spec_schema, data, version='1.0')
        ex = self.assertRaises(exc.ESchema, spec.validate)
        msg = 'key1 (min_version=1.1) is not supported by spec version 1.0.'
        self.assertIn(msg, six.text_type(ex.message))

    def test_validate_version_fail_version_over_max(self):
        spec_schema = {
            'type': schema.String('Type name', required=True),
            'version': schema.String('Version number', required=True),
            'key1': schema.String('first key', default='value1',
                                  max_version='2.0'),
            'key2': schema.Integer('second key', required=True),
        }

        data = {
            'key1': 'value1',
            'key2': 2,
            'type': 'test-type',
            'version': '3.0'
        }
        spec = schema.Spec(spec_schema, data, version='3.0')
        ex = self.assertRaises(exc.ESchema, spec.validate)
        msg = 'key1 (max_version=2.0) is not supported by spec version 3.0.'
        self.assertIn(msg, six.text_type(ex.message))

    def test_resolve_value(self):
        data = {'key2': 2}
        sot = schema.Spec(self.spec_schema, data, version='1.2')

        res = sot.resolve_value('key2')
        self.assertEqual(2, res)

        res = sot.resolve_value('key1')
        self.assertEqual('value1', res)

        ex = self.assertRaises(exc.ESchema, sot.resolve_value, 'key3')
        self.assertEqual("Invalid spec item: key3", six.text_type(ex))

    def test_resolve_value_required_key_missing(self):
        data = {'key1': 'value1'}
        sot = schema.Spec(self.spec_schema, data, version='1.0')

        ex = self.assertRaises(exc.ESchema, sot.resolve_value, 'key2')
        self.assertIn("Required spec item 'key2' not provided",
                      six.text_type(ex.message))

    def test___getitem__(self):
        data = {'key2': 2}
        sot = schema.Spec(self.spec_schema, data, version='1.2')

        res = sot['key1']
        self.assertEqual('value1', res)
        res = sot['key2']
        self.assertEqual(2, res)

    def test___len__(self):
        data = {'key2': 2}
        sot = schema.Spec(self.spec_schema, data, version='1.2')

        res = len(sot)
        self.assertEqual(2, res)

    def test___contains__(self):
        data = {'key2': 2}
        sot = schema.Spec(self.spec_schema, data, version='1.2')

        self.assertIn('key1', sot)
        self.assertIn('key2', sot)
        self.assertNotIn('key3', sot)

    def test__iter__(self):
        data = {'key2': 2}
        sot = schema.Spec(self.spec_schema, data, version='1.2')

        res = [k for k in iter(sot)]

        self.assertIn('key1', res)
        self.assertIn('key2', res)


class TestSpecVersionChecking(base.SenlinTestCase):

    def test_spec_version_okay(self):
        spec = {'type': 'Foo', 'version': 'version string'}
        res = schema.get_spec_version(spec)
        self.assertEqual(('Foo', 'version string'), res)

        spec = {'type': 'Foo', 'version': 1.5}
        res = schema.get_spec_version(spec)
        self.assertEqual(('Foo', '1.5'), res)

    def test_spec_version_not_dict(self):
        spec = 'a string'
        ex = self.assertRaises(exc.ESchema, schema.get_spec_version, spec)
        self.assertEqual('The provided spec is not a map.',
                         six.text_type(ex))

    def test_spec_version_no_type_key(self):
        spec = {'tpye': 'a string'}
        ex = self.assertRaises(exc.ESchema, schema.get_spec_version, spec)
        self.assertEqual("The 'type' key is missing from the provided "
                         "spec map.", six.text_type(ex))

    def test_spec_version_no_version_key(self):
        spec = {'type': 'a string', 'ver': '123'}
        ex = self.assertRaises(exc.ESchema, schema.get_spec_version, spec)
        self.assertEqual("The 'version' key is missing from the provided "
                         "spec map.", six.text_type(ex))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_utils import reflection


class EventBackend(object):

    @classmethod
    def _check_entity(cls, e):
        e_type = reflection.get_class_name(e, fully_qualified=False)
        return e_type.upper()

    @classmethod
    def _get_action_name(cls, action):
        """Get action name by inference.

        :param action: An action object.
        :returns: A string containing the inferred action name.
        """
        name = action.action.split('_', 1)
        if len(name) == 1:
            return name[0].lower()

        name = name[1].lower()
        if name == "operation":
            name = action.inputs.get("operation", name)
        return name

    @classmethod
    def dump(cls, level, action, **kwargs):
        """A method for sub-class to override.

        :param level: An integer as defined by python logging module.
        :param action: The action that triggered this dump.
        :param dict kwargs: Additional parameters such as ``phase``,
                            ``timestamp`` or ``extra``.
        :returns: None
        """
        raise NotImplementedError



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from senlin.common import context
from senlin.tests.unit.common import base


class TestRequestContext(base.SenlinTestCase):

    def setUp(self):
        self.ctx = {
            'auth_url': 'http://xyz',
            'auth_token_info': {'123info': 'woop'},
            'user_name': 'mick',
            'user_domain_name': 'user-domain-name',
            'project_id': 'project-id',
            'project_name': 'a project',
            'project_domain_name': 'a project domain',
            'domain_name': 'this domain',
            'trusts': None,
            'region_name': 'regionOne',
            'password': 'foo',
            'is_admin': False  # needed for tests to work
        }

        super(TestRequestContext, self).setUp()

    def test_request_context_init(self):
        ctx = context.RequestContext(
            auth_url=self.ctx.get('auth_url'),
            auth_token_info=self.ctx.get('auth_token_info'),
            user_name=self.ctx.get('user_name'),
            user_domain_name=self.ctx.get('user_domain_name'),
            project_id=self.ctx.get('project_id'),
            project_name=self.ctx.get('project_name'),
            project_domain_name=self.ctx.get('project_domain_name'),
            domain_name=self.ctx.get('domain_name'),
            trusts=self.ctx.get('trusts'),
            region_name=self.ctx.get('region_name'),
            password=self.ctx.get('password'),
            is_admin=self.ctx.get('is_admin'))  # need for tests to work

        ctx_dict = ctx.to_dict()
        for k, v in self.ctx.items():
            self.assertEqual(v, ctx_dict.get(k))

    def test_request_context_from_dict(self):
        ctx = context.RequestContext.from_dict(self.ctx)

        ctx_dict = ctx.to_dict()
        for k, v in self.ctx.items():
            self.assertEqual(v, ctx_dict.get(k))

    def test_request_context_update(self):
        ctx = context.RequestContext.from_dict(self.ctx)

        for k in self.ctx:
            self.assertEqual(self.ctx.get(k), ctx.to_dict().get(k))
            override = '%s_override' % k
            setattr(ctx, k, override)
            self.assertEqual(override, ctx.to_dict().get(k))

    def test_get_admin_context(self):
        ctx1 = context.get_admin_context()
        self.assertTrue(ctx1.is_admin)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Health registry object."""

from senlin.db import api as db_api
from senlin.objects import base
from senlin.objects import fields


@base.SenlinObjectRegistry.register
class HealthRegistry(base.SenlinObject, base.VersionedObjectDictCompat):
    """Senlin health registry object."""

    fields = {
        'id': fields.UUIDField(),
        'cluster_id': fields.UUIDField(),
        'check_type': fields.StringField(),
        'interval': fields.IntegerField(nullable=True),
        'params': fields.JsonField(nullable=True),
        'engine_id': fields.UUIDField(),
        'enabled': fields.BooleanField(),
    }

    @classmethod
    def create(cls, context, cluster_id, check_type, interval, params,
               engine_id, enabled=True):
        obj = db_api.registry_create(context, cluster_id, check_type,
                                     interval,
                                     params, engine_id,
                                     enabled=enabled)
        return cls._from_db_object(context, cls(), obj)

    @classmethod
    def update(cls, context, cluster_id, values):
        db_api.registry_update(context, cluster_id, values)

    @classmethod
    def claim(cls, context, engine_id):
        objs = db_api.registry_claim(context, engine_id)
        return [cls._from_db_object(context, cls(), obj) for obj in objs]

    @classmethod
    def delete(cls, context, cluster_id):
        db_api.registry_delete(context, cluster_id)

    @classmethod
    def get(cls, context, cluster_id):
        obj = db_api.registry_get(context, cluster_id)
        return cls._from_db_object(context, cls(), obj)

    @classmethod
    def get_by_engine(cls, context, engine_id, cluster_id):
        params = {
            "cluster_id": cluster_id,
            "engine_id": engine_id,
        }
        obj = db_api.registry_get_by_param(context, params)
        return cls._from_db_object(context, cls(), obj)

    @classmethod
    def disable_registry(cls, context, cluster_id):
        cls.update(context, cluster_id, {'enabled': False})

    @classmethod
    def enable_registry(cls, context, cluster_id):
        cls.update(context, cluster_id, {"enabled": True})



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_utils import versionutils

from senlin.common import consts
from senlin.objects import base
from senlin.objects import fields

CONF = cfg.CONF
CONF.import_opt('default_action_timeout', 'senlin.common.config')


@base.SenlinObjectRegistry.register
class ClusterListRequest(base.SenlinObject):

    fields = {
        'name': fields.ListOfStringsField(nullable=True),
        'status': fields.ListOfEnumField(
            valid_values=list(consts.CLUSTER_STATUSES), nullable=True),
        'limit': fields.NonNegativeIntegerField(nullable=True),
        'marker': fields.UUIDField(nullable=True),
        'sort': fields.SortField(
            valid_keys=list(consts.CLUSTER_SORT_KEYS), nullable=True),
        'project_safe': fields.FlexibleBooleanField(default=True),
    }


@base.SenlinObjectRegistry.register
class ClusterCreateRequestBody(base.SenlinObject):

    # VERSION 1.0: initial version
    # VERSION 1.1: added field 'config'
    VERSION = '1.1'
    VERSION_MAP = {
        '1.7': '1.1',
    }

    fields = {
        'name': fields.NameField(),
        'profile_id': fields.StringField(),
        'min_size': fields.CapacityField(
            nullable=True, minimum=0,
            default=consts.CLUSTER_DEFAULT_MIN_SIZE),
        'max_size': fields.CapacityField(
            nullable=True, minimum=-1,
            default=consts.CLUSTER_DEFAULT_MAX_SIZE),
        'desired_capacity': fields.CapacityField(
            nullable=True, minimum=0),
        'metadata': fields.JsonField(nullable=True, default={}),
        'timeout': fields.NonNegativeIntegerField(
            nullable=True, default=CONF.default_action_timeout),
        'config': fields.JsonField(nullable=True, default={}),
    }

    def obj_make_compatible(self, primitive, target_version):
        super(ClusterCreateRequest, self).obj_make_compatible(
            primitive, target_version)
        target_version = versionutils.convert_version_to_tuple(target_version)
        if target_version < (1, 1):
            if 'config' in primitive['senlin_object.data']:
                del primitive['senlin_object.data']['config']


@base.SenlinObjectRegistry.register
class ClusterCreateRequest(base.SenlinObject):

    fields = {
        'cluster': fields.ObjectField('ClusterCreateRequestBody')
    }


@base.SenlinObjectRegistry.register
class ClusterGetRequest(base.SenlinObject):

    fields = {
        'identity': fields.StringField()
    }


@base.SenlinObjectRegistry.register
class ClusterUpdateRequest(base.SenlinObject):

    # VERSION 1.0: initial version
    # VERSION 1.1: added field 'profile_only'
    # VERSION 1.2: added field 'config'
    VERSION = '1.2'
    VERSION_MAP = {
        '1.6': '1.1',
        '1.7': '1.2',
    }

    fields = {
        'identity': fields.StringField(),
        'name': fields.NameField(nullable=True),
        'profile_id': fields.StringField(nullable=True),
        'metadata': fields.JsonField(nullable=True),
        'timeout': fields.NonNegativeIntegerField(nullable=True),
        'profile_only': fields.BooleanField(nullable=True),
        'config': fields.JsonField(nullable=True),
    }

    def obj_make_compatible(self, primitive, target_version):
        super(ClusterUpdateRequest, self).obj_make_compatible(
            primitive, target_version)
        target_version = versionutils.convert_version_to_tuple(target_version)
        if target_version < (1, 1):
            if 'profile_only' in primitive['senlin_object.data']:
                del primitive['senlin_object.data']['profile_only']
        if target_version < (1, 2):
            if 'config' in primitive['senlin_object.data']:
                del primitive['senlin_object.data']['config']


@base.SenlinObjectRegistry.register
class ClusterAddNodesRequest(base.SenlinObject):

    fields = {
        'identity': fields.StringField(),
        'nodes': fields.IdentityListField(min_items=1)
    }


@base.SenlinObjectRegistry.register
class ClusterDelNodesRequest(base.SenlinObject):

    # VERSION 1.0: Initial version
    # VERSION 1.1: Add field 'destroy_after_deletion'
    VERSION = '1.1'
    VERSION_MAP = {
        '1.4': '1.1',
    }

    fields = {
        'identity': fields.StringField(),
        'nodes': fields.IdentityListField(min_items=1),
        'destroy_after_deletion': fields.BooleanField(nullable=True,
                                                      default=False)
    }

    def obj_make_compatible(self, primitive, target_version):
        super(ClusterDelNodesRequest, self).obj_make_compatible(
            primitive, target_version)
        target_version = versionutils.convert_version_to_tuple(target_version)
        if target_version < (1, 1):
            if 'destroy_after_deletion' in primitive['senlin_object.data']:
                del primitive['senlin_object.data']['destroy_after_deletion']


@base.SenlinObjectRegistry.register
class ClusterReplaceNodesRequest(base.SenlinObject):

    fields = {
        'identity': fields.StringField(),
        'nodes': fields.NodeReplaceMapField(),
    }


@base.SenlinObjectRegistry.register
class ClusterResizeRequest(base.SenlinObject):

    fields = {
        'identity': fields.StringField(),
        'adjustment_type': fields.AdjustmentTypeField(nullable=True),
        'number': fields.FloatField(nullable=True),
        'min_size': fields.CapacityField(nullable=True, minimum=0),
        'max_size': fields.CapacityField(nullable=True, minimum=-1),
        'min_step': fields.NonNegativeIntegerField(nullable=True),
        'strict': fields.BooleanField(nullable=True, default=True),
    }


@base.SenlinObjectRegistry.register
class ClusterScaleInRequest(base.SenlinObject):

    fields = {
        'identity': fields.StringField(),
        'count': fields.NonNegativeIntegerField(nullable=True),
    }


@base.SenlinObjectRegistry.register
class ClusterScaleOutRequest(base.SenlinObject):

    fields = {
        'identity': fields.StringField(),
        'count': fields.NonNegativeIntegerField(nullable=True),
    }


@base.SenlinObjectRegistry.register
class ClusterAttachPolicyRequest(base.SenlinObject):

    fields = {
        'identity': fields.StringField(),
        'policy_id': fields.StringField(),
        'enabled': fields.BooleanField(nullable=True, default=True),
    }


@base.SenlinObjectRegistry.register
class ClusterUpdatePolicyRequest(base.SenlinObject):

    fields = {
        'identity': fields.StringField(),
        'policy_id': fields.StringField(),
        'enabled': fields.BooleanField(nullable=True, default=True),
    }


@base.SenlinObjectRegistry.register
class ClusterDetachPolicyRequest(base.SenlinObject):

    fields = {
        'identity': fields.StringField(),
        'policy_id': fields.StringField(),
    }


@base.SenlinObjectRegistry.register
class ClusterCheckRequest(base.SenlinObject):

    fields = {
        'identity': fields.StringField(),
        'params': fields.JsonField(nullable=True),
    }


@base.SenlinObjectRegistry.register
class ClusterRecoverRequest(base.SenlinObject):

    fields = {
        'identity': fields.StringField(),
        'params': fields.JsonField(nullable=True),
    }


@base.SenlinObjectRegistry.register
class ClusterCollectRequest(base.SenlinObject):

    fields = {
        'identity': fields.StringField(),
        'path': fields.StringField(),
    }


@base.SenlinObjectRegistry.register
class ClusterOperationRequest(base.SenlinObject):

    fields = {
        'identity': fields.StringField(),
        'operation': fields.StringField(),
        'filters': fields.JsonField(nullable=True, default={}),
        'params': fields.JsonField(nullable=True, default={}),
    }


@base.SenlinObjectRegistry.register
class ClusterDeleteRequest(base.SenlinObject):
    # VERSION 1.0: Initial version
    # VERSION 1.1 Added field 'force'
    VERSION = '1.1'
    VERSION_MAP = {
        '1.8': '1.1',
    }

    fields = {
        'identity': fields.StringField(),
        'force': fields.BooleanField(default=False)
    }

    def obj_make_compatible(self, primitive, target_version):
        super(ClusterDeleteRequest, self).obj_make_compatible(
            primitive, target_version)
        target_version = versionutils.convert_version_to_tuple(target_version)
        if target_version < (1, 1):
            if 'force' in primitive['senlin_object.data']:
                del primitive['senlin_object.data']['force']


@base.SenlinObjectRegistry.register
class ClusterCompleteLifecycleRequest(base.SenlinObject):

    fields = {
        'identity': fields.StringField(),
        'lifecycle_action_token': fields.StringField(),
    }



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import six
import testtools

from senlin.common import constraints
from senlin.common import exception as exc
from senlin.common import schema


class TestConstraintsSchema(testtools.TestCase):
    def test_allowed_values(self):
        d = {
            'constraint': ['foo', 'bar'],
            'type': 'AllowedValues'
        }
        r = constraints.AllowedValues(['foo', 'bar'])

        self.assertEqual(d, dict(r))

    def test_allowed_values_numeric_int(self):
        """Test AllowedValues constraint for numeric integer values.

        Test if the AllowedValues constraint works for numeric values in any
        combination of numeric strings or numbers in the constraint and
        numeric strings or numbers as value.
        """

        # Allowed values defined as integer numbers
        s = schema.Integer(
            constraints=[constraints.AllowedValues([1, 2, 4])]
        )
        # ... and value as number or string
        self.assertIsNone(s.validate(1))

        err = self.assertRaises(exc.ESchema, s.validate, 3)
        self.assertEqual("'3' must be one of the allowed values: 1, 2, 4",
                         six.text_type(err))

        self.assertIsNone(s.validate('1'))
        err = self.assertRaises(exc.ESchema, s.validate, '3')
        self.assertEqual("'3' must be one of the allowed values: 1, 2, 4",
                         six.text_type(err))

        # Allowed values defined as integer strings
        s = schema.Integer(
            constraints=[constraints.AllowedValues(['1', '2', '4'])]
        )
        # ... and value as number or string
        self.assertIsNone(s.validate(1))
        err = self.assertRaises(exc.ESchema, s.validate, 3)
        self.assertEqual("'3' must be one of the allowed values: 1, 2, 4",
                         six.text_type(err))

        self.assertIsNone(s.validate('1'))
        err = self.assertRaises(exc.ESchema, s.validate, '3')
        self.assertEqual("'3' must be one of the allowed values: 1, 2, 4",
                         six.text_type(err))

    def test_allowed_values_numeric_float(self):
        """Test AllowedValues constraint for numeric floating point values.

        Test if the AllowedValues constraint works for numeric values in any
        combination of numeric strings or numbers in the constraint and
        numeric strings or numbers as value.
        """

        # Allowed values defined as numbers
        s = schema.Number(
            constraints=[constraints.AllowedValues([1.1, 2.2, 4.4])]
        )
        # ... and value as number or string
        self.assertIsNone(s.validate_constraints(1.1))
        err = self.assertRaises(exc.ESchema, s.validate_constraints, 3.3)
        self.assertEqual("'3.3' must be one of the allowed values: "
                         "1.1, 2.2, 4.4", six.text_type(err))
        self.assertIsNone(s.validate_constraints('1.1', s))
        err = self.assertRaises(exc.ESchema, s.validate_constraints, '3.3')
        self.assertEqual("'3.3' must be one of the allowed values: "
                         "1.1, 2.2, 4.4", six.text_type(err))

        # Allowed values defined as strings
        s = schema.Number(
            constraints=[constraints.AllowedValues(['1.1', '2.2', '4.4'])]
        )
        # ... and value as number or string
        self.assertIsNone(s.validate_constraints(1.1, s))
        err = self.assertRaises(exc.ESchema, s.validate_constraints, 3.3, s)
        self.assertEqual("'3.3' must be one of the allowed values: "
                         "1.1, 2.2, 4.4", six.text_type(err))
        self.assertIsNone(s.validate_constraints('1.1', s))
        err = self.assertRaises(exc.ESchema, s.validate_constraints, '3.3', s)
        self.assertEqual("'3.3' must be one of the allowed values: "
                         "1.1, 2.2, 4.4", six.text_type(err))

    def test_schema_all(self):
        d = {
            'type': 'String',
            'description': 'A string',
            'default': 'wibble',
            'required': True,
            'updatable': False,
            'constraints': [{
                'constraint': ['foo', 'bar'],
                'type': 'AllowedValues'
            }]
        }
        c = constraints.AllowedValues(['foo', 'bar'])
        s = schema.String('A string', default='wibble', required=True,
                          constraints=[c])
        self.assertEqual(d, dict(s))

    def test_schema_list_schema(self):
        d = {
            'type': 'List',
            'description': 'A list',
            'schema': {
                '*': {
                    'type': 'String',
                    'description': 'A string',
                    'default': 'wibble',
                    'required': True,
                    'updatable': False,
                    'constraints': [{
                        'constraint': ['foo', 'bar'],
                        'type': 'AllowedValues'
                    }]
                }
            },
            'required': False,
            'updatable': False,
        }
        c = constraints.AllowedValues(['foo', 'bar'])
        s = schema.String('A string', default='wibble', required=True,
                          constraints=[c])
        l = schema.List('A list', schema=s)
        self.assertEqual(d, dict(l))

    def test_schema_map_schema(self):
        d = {
            'type': 'Map',
            'description': 'A map',
            'schema': {
                'Foo': {
                    'type': 'String',
                    'description': 'A string',
                    'default': 'wibble',
                    'required': True,
                    'updatable': False,
                    'constraints': [{
                        'type': 'AllowedValues',
                        'constraint': ['foo', 'bar']
                    }]
                }
            },
            'required': False,
            'updatable': False,
        }
        c = constraints.AllowedValues(['foo', 'bar'])
        s = schema.String('A string', default='wibble', required=True,
                          constraints=[c])
        m = schema.Map('A map', schema={'Foo': s})
        self.assertEqual(d, dict(m))

    def test_schema_nested_schema(self):
        d = {
            'type': 'List',
            'description': 'A list',
            'schema': {
                '*': {
                    'type': 'Map',
                    'description': 'A map',
                    'schema': {
                        'Foo': {
                            'type': 'String',
                            'description': 'A string',
                            'default': 'wibble',
                            'required': True,
                            'updatable': False,
                            'constraints': [{
                                'type': 'AllowedValues',
                                'constraint': ['foo', 'bar']
                            }]
                        }
                    },
                    'required': False,
                    'updatable': False,
                }
            },
            'required': False,
            'updatable': False,
        }
        c = constraints.AllowedValues(['foo', 'bar'])
        s = schema.String('A string', default='wibble', required=True,
                          constraints=[c])
        m = schema.Map('A map', schema={'Foo': s})
        l = schema.List('A list', schema=m)
        self.assertEqual(d, dict(l))

    def test_schema_validate_good(self):
        c = constraints.AllowedValues(['foo', 'bar'])
        s = schema.String('A string', default='wibble', required=True,
                          constraints=[c])
        self.assertIsNone(s.validate('foo'))

    def test_schema_validate_fail(self):
        c = constraints.AllowedValues(['foo', 'bar'])
        s = schema.String('A string', default='wibble', required=True,
                          constraints=[c])
        err = self.assertRaises(exc.ESchema, s.validate, 'zoo')
        self.assertIn("'zoo' must be one of the allowed values: foo, bar",
                      six.text_type(err))

    def test_schema_nested_validate_good(self):
        c = constraints.AllowedValues(['foo', 'bar'])
        nested = schema.String('A string', default='wibble', required=True,
                               constraints=[c])
        s = schema.Map('A map', schema={'Foo': nested})
        self.assertIsNone(s.validate({'Foo': 'foo'}))

    def test_schema_nested_validate_fail(self):
        c = constraints.AllowedValues(['foo', 'bar'])
        nested = schema.String('A string', default='wibble', required=True,
                               constraints=[c])
        s = schema.Map('A map', schema={'Foo': nested})
        err = self.assertRaises(exc.ESchema, s.validate, {'Foo': 'zoo'})

        self.assertIn("'zoo' must be one of the allowed values: foo, bar",
                      six.text_type(err))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import testtools

from senlin.db import api as db_api
from senlin.objects import base
from senlin.objects import health_registry as hro


class TestHealthRegistry(testtools.TestCase):

    def setUp(self):
        super(TestHealthRegistry, self).setUp()
        self.ctx = mock.Mock()

    @mock.patch.object(base.SenlinObject, '_from_db_object')
    @mock.patch.object(db_api, 'registry_create')
    def test_create(self, mock_create, mock_from):
        x_registry = mock.Mock()
        mock_create.return_value = x_registry
        x_obj = mock.Mock()
        mock_from.return_value = x_obj

        result = hro.HealthRegistry.create(
            self.ctx, "FAKE_ID", "FAKE_TYPE", 123, {'foo': 'bar'},
            'FAKE_ENGINE')

        self.assertEqual(x_obj, result)
        mock_create.assert_called_once_with(
            self.ctx, "FAKE_ID", "FAKE_TYPE", 123, {'foo': 'bar'},
            "FAKE_ENGINE", enabled=True)
        mock_from.assert_called_once_with(self.ctx, mock.ANY, x_registry)

    @mock.patch.object(db_api, 'registry_update')
    def test_update(self, mock_update):
        hro.HealthRegistry.update(self.ctx, "FAKE_ID", {"foo": "bar"})

        mock_update.assert_called_once_with(
            self.ctx, "FAKE_ID", {"foo": "bar"})

    @mock.patch.object(base.SenlinObject, '_from_db_object')
    @mock.patch.object(db_api, 'registry_claim')
    def test_claim(self, mock_claim, mock_from):
        x_registry = mock.Mock()
        mock_claim.return_value = [x_registry]
        x_obj = mock.Mock()
        mock_from.side_effect = [x_obj]

        result = hro.HealthRegistry.claim(self.ctx, "FAKE_ENGINE")

        self.assertEqual([x_obj], result)
        mock_claim.assert_called_once_with(self.ctx, "FAKE_ENGINE")
        mock_from.assert_called_once_with(self.ctx, mock.ANY, x_registry)

    @mock.patch.object(db_api, 'registry_delete')
    def test_delete(self, mock_delete):
        hro.HealthRegistry.delete(self.ctx, "FAKE_ID")

        mock_delete.assert_called_once_with(self.ctx, "FAKE_ID")

    @mock.patch.object(base.SenlinObject, '_from_db_object')
    @mock.patch.object(db_api, 'registry_get')
    def test_get(self, mock_get, mock_from):
        x_registry = mock.Mock()
        x_registry.cluster_id = 'FAKE'
        mock_get.return_value = x_registry

        x_obj = mock.Mock()
        mock_from.return_value = x_obj

        result = hro.HealthRegistry.get(self.ctx, 'FAKE')

        self.assertEqual(x_obj, result)
        mock_get.assert_called_once_with(self.ctx, 'FAKE')
        mock_from.assert_called_once_with(self.ctx, mock.ANY, x_registry)

    @mock.patch.object(base.SenlinObject, '_from_db_object')
    @mock.patch.object(db_api, 'registry_get_by_param')
    def test_get_by_engine(self, mock_get, mock_from):
        x_registry = mock.Mock()
        x_registry.cluster_id = 'FAKE'
        x_registry.engine_id = 'FAKE_ENGINE'
        mock_get.return_value = x_registry

        x_obj = mock.Mock()
        mock_from.return_value = x_obj

        result = hro.HealthRegistry.get_by_engine(
            self.ctx, 'FAKE_ENGINE', 'FAKE')

        self.assertEqual(x_obj, result)
        mock_get.assert_called_once_with(
            self.ctx, {"cluster_id": "FAKE", "engine_id": "FAKE_ENGINE"})
        mock_from.assert_called_once_with(self.ctx, mock.ANY, x_registry)

    @mock.patch.object(hro.HealthRegistry, 'update')
    def test_disable(self, mock_update):
        hro.HealthRegistry.disable_registry(
            self.ctx, "FAKE_ID")

        mock_update.assert_called_once_with(
            self.ctx, "FAKE_ID", {"enabled": False})

    @mock.patch.object(hro.HealthRegistry, 'update')
    def test_enable(self, mock_update):
        hro.HealthRegistry.enable_registry(
            self.ctx, "FAKE_ID")

        mock_update.assert_called_once_with(
            self.ctx, "FAKE_ID", {"enabled": True})



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from senlin.api.common import util
from senlin.api.common import wsgi
from senlin.common import context
from senlin.common import exception
from senlin.drivers import base as driver_base
from senlin.rpc import client as rpc


class TrustMiddleware(wsgi.Middleware):
    """Extract trust info from request.

    The extracted information is filled into the request context.
    Senlin engine will use this information for access control.
    """
    def _get_trust(self, req):
        """List trusts with current user as the trustor.

        :param req: The WSGI request object.
        :return: ID of the trust or exception of InternalError.
        """
        rpcc = rpc.EngineClient()

        ctx = req.context
        params = {'user': ctx.user_id, 'project': ctx.project_id}
        obj = util.parse_request('CredentialGetRequest', req, params)
        res = rpcc.call(ctx, 'credential_get', obj)
        if res:
            trust_id = res.get('trust', None)
            if trust_id:
                return trust_id

        params = {
            'auth_url': ctx.auth_url,
            'token': ctx.auth_token,
            'project_id': ctx.project_id,
            'user_id': ctx.user_id,
        }
        kc = driver_base.SenlinDriver().identity(params)
        service_cred = context.get_service_credentials()
        admin_id = kc.get_user_id(**service_cred)
        try:
            trust = kc.trust_get_by_trustor(ctx.user_id, admin_id,
                                            ctx.project_id)
        except exception.InternalError as ex:
            if ex.code == 400:
                trust = None
            else:
                raise
        if not trust:
            # Create a trust if no existing one found
            trust = kc.trust_create(ctx.user_id, admin_id, ctx.project_id,
                                    ctx.roles)

        # If credential not exists, create it, otherwise update it.
        cred = {'openstack': {'trust': trust.id}}
        params = {'cred': cred}
        obj = util.parse_request('CredentialCreateRequest', req, params)
        rpcc.call(ctx, 'credential_create', obj)

        return trust.id

    def process_request(self, req):
        trust_id = self._get_trust(req)
        req.context.trusts = trust_id
